# -*- coding: utf-8 -*-
# themes.py
"""This module provides a graphic theme for plotting"""

# Copyright (C) 2010-2015 Federico Brega, Pierluigi Villani

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.


class ThemeManager:
    def getthemeslist(self):
        """Return names associated with classes subclassing Theme."""
        return sorted(Theme.themes.keys())

    def gettheme(self, name):
        """Return reference to the given class name."""
        if name in Theme.themes:
            return Theme.themes[name]
        return None


class ThemeMount(type):  # from http://martyalchin.com/2008/jan/10/simple-plugin-framework/
    def __init__(cls, name, bases, attrs):
        if not hasattr(cls, 'themes'):
            # This branch only executes when processing the mount point itself.
            # So, since this is a new plugin type, not an implementation, this
            # class shouldn't be registered as a plugin. Instead, it sets up a
            # list where plugins can be registered later.
            cls.themes = {}
        else:
            # This must be a plugin implementation, which should be registered.
            # Simply appending it to the list is all that's needed to keep
            # track of it later.
            # if cls.themename in cls.themes:
                # TODO: if user can create/install themes warn if a name is
                # used by more than one theme
                # return
            cls.themes[cls.themename] = cls


class Theme(metaclass=ThemeMount):
    """Model of a theme: every theme must subclass this class."""
    themename = ""
    shear = -0.15

    @staticmethod
    def getcolor(colors, levels, grad):
        """Return gradient relative color"""
        pass

    @staticmethod
    def drawslopeinfo(devc, info, settings, startx, starty):
        """Draw slope's name and other info background"""
        pass

    @staticmethod
    def drawlegend(devc, settings, startx, starty, title):
        """Draw slope's legend"""
        pass

    @staticmethod
    def paintbackground(devc, width, height):
        """Draw background"""
        pass

    @staticmethod
    def gradback(devc, width, height, setting_f, startcol="rgb(64, 185, 255)", endcol='white'):
        """Draw a gradient background"""
        pass

    @staticmethod
    def olines(devc, slope, margins, width, height, metersize):
        """Draw altitude line"""
        pass

    @staticmethod
    def alttext(devc, slope, margins, max_y, metersize):
        """draw altitude text aside"""
        pass

    @staticmethod
    def yaxis(devc, slope, margins, max_y, metersize):
        """Draw y axis"""
        pass

    @staticmethod
    def xaxis(devc, slope, margins, max_y, d_incr, kmsize, dx, dy):
        """Draw km bar"""
        pass

    @staticmethod
    def fillhslopecontour(devc, spath_back, dx, dy, colorlist):
        """ Draw horizontal contour."""
        pass

    @staticmethod
    def fillhpoly(devc, points, color):
        """fill horizontal polygons of the 3d slope"""
        pass

    @staticmethod
    def fillfirsthpoly(devc, points, color):
        """fill first horizontal polygon of the 3d slope"""

    @staticmethod
    def fillvslopecontour(devc, path_points, pixel_0m, pixel_1000m, vpolygons, colorlist):
        """Draw a close smooth path thant contais the whole slope."""
        pass

    @staticmethod
    def fillvpoly(devc, points, color):
        """fill vertical polygons (2d and 3d) of the slope"""
        pass

    @staticmethod
    def desctext(devc, text, pos_x, pos_y, font):
        """Draw description text for every cp"""
        pass

    @staticmethod
    def gradtext(devc, text, pos_x, pos_y, font):
        """Draw gradient text for every cp"""
        pass


class ThemeClassic(Theme):
    """Classic CycloGraph theme"""
    themename = "Classic"
    shear = -0.15

    @staticmethod
    def getcolor(colors, levels, grad):
        """Return gradient relative color"""
        gradient = abs(grad)
        for i in range(len(levels)):
            if gradient < levels[i]:
                return colors[i-1]
        return colors[i]

    @staticmethod
    def drawslopeinfo(devc, info, settings, startx, starty):
        """Draw slope's name and other info background"""
        name, avp, maxp, eleg, heightg = info
        devc.setfont(settings['ftitle'])
        devc.drawtext(name, startx, starty)
        if settings['sinfo']:
            devc.setfont(settings['fdesc'])
            devc.drawtext(avp, startx+10, starty+20)
            devc.drawtext(maxp, startx+10, starty+40)
            devc.drawtext(eleg, startx+10, starty+60)
            devc.drawtext(heightg, startx+10, starty+80)

    @staticmethod
    def drawlegend(devc, settings, startx, starty, title):
        """Draw slope's legend"""
        devc.setbrush("rgb(236, 236, 236)")
        rec = (startx, starty, 310, 52)
        devc.drawrectangle(*rec)
        devc.setfont(settings['fdesc'])
        devc.setpen('black', 2)
        devc.drawtext(title, startx+130, starty+2)
        incr = 43
        devc.setpen('black', 1)
        for i in range(len(settings['colors'])):
            devc.setbrush(settings['colors'][i])
            rec = (startx+8+i*incr, starty+20, 36, 10)
            devc.drawrectangle(*rec)
            alti_t = (">%d%%" % (settings['levels'][i]))
            devc.drawtext(alti_t, startx+9+i*incr, starty+34)

    @staticmethod
    def paintbackground(devc, width, height):
        """Draw backgroundr"""
        devc.setbrush("rgb(255, 255, 255)")
        rec = (0, 0, width, height)
        devc.drawrectangle(*rec)

    @staticmethod
    def gradback(devc, width, height, setting_f, startcol="rgb(64, 185, 255)", endcol='white'):
        """Draw a gradient background"""
        area = (0, 0, width, height)
        devc.gradfill(area, startcol, endcol)
        devc.setpen('black', 1)
        devc.setfont(setting_f)

    @staticmethod
    def olines(devc, slope, margins, width, height, metersize):
        """Draw altitude line"""
        devc.setpen("rgb(150, 150, 150)", 1)
        (updownmar, leftrightmar) = margins
        (lef_mar, rig_mar) = leftrightmar
        (upp_mar, low_mar) = updownmar
        x0_alt_bar = lef_mar - 3
        x1_alt_bar = width - rig_mar - 3
        y_alt_base_bar = height - low_mar
        for i in range(int(slope.min_h / slope.h_incr),
                       int((slope.max_h / slope.h_incr) + 1)):
            yincr = int((i * slope.h_incr - slope.min_h) * metersize)
            devc.drawline(x0_alt_bar, y_alt_base_bar - yincr,
                          x1_alt_bar, y_alt_base_bar - yincr)

    @staticmethod
    def alttext(devc, slope, margins, max_y, metersize):
        """draw altitude text aside"""
        (updownmar, leftrightmar) = margins
        (lef_mar, rig_mar) = leftrightmar
        (upp_mar, low_mar) = updownmar

        devc.setpen('black', 2)
        x_alt_txt = lef_mar - 40
        y_alt_base_txt = max_y - low_mar - 8
        for i in range(int(slope.min_h/slope.h_incr), int((slope.max_h/slope.h_incr) + 1)):
            yincr = int((i * slope.h_incr - slope.min_h) * metersize)
            alti_t = ("%4d" % (i*slope.h_incr))
            devc.drawtext(alti_t, x_alt_txt, y_alt_base_txt - yincr)

    @staticmethod
    def yaxis(devc, slope, margins, max_y, metersize):
        """Draw y axis"""
        (updownmar, leftrightmar) = margins
        (lef_mar, rig_mar) = leftrightmar
        (upp_mar, low_mar) = updownmar
        linesnum = int(slope.max_h / slope.h_incr)
        devc.drawline(lef_mar,  max_y - low_mar + 4,
                      lef_mar,  max_y - low_mar - 8 \
                      - int(linesnum * slope.h_incr - slope.min_h) * metersize)

    @staticmethod
    def xaxis(devc, slope, margins, max_y, d_incr, kmsize, dx, dy):
        """Draw km bar"""
        (updownmar, leftrightmar) = margins
        (lef_mar, rig_mar) = leftrightmar
        (upp_mar, low_mar) = updownmar
        for i in range(int(int(slope.max_d) / d_incr - int(slope.min_d) / d_incr + 1)):
            if i % 2 == 0:
                devc.setbrush("rgb(0, 0, 0)")
            else:
                devc.setbrush("rgb(255, 255, 255)")
            rec = (i * d_incr * kmsize + lef_mar + dx,
                   max_y - low_mar + dy, d_incr * kmsize + 1, 5)
            devc.drawrectangle(*rec)
            d_label = ("%d" % (i * d_incr + int(slope.min_d)))
            devc.drawtext(d_label, rec[0], rec[1] + 5)

    @staticmethod
    def fillhpoly(devc, points, color):
        """fill horizontal polygons of the 3d slope"""
        devc.setbrush(color)
        devc.drawpolygon(points)

    @staticmethod
    def fillfirsthpoly(devc, points, color):
        """fill first horizontal polygon of the 3d slope"""
        devc.setbrush(color)
        devc.drawpolygon(points)

    @staticmethod
    def fillvpoly(devc, points, color):
        """fill vertical polygons (2d and 3d) of the slope"""
        devc.setbrush(color)
        devc.drawpolygon(points)

    @staticmethod
    def desctext(devc, text, pos_x, pos_y, font):
        """Draw description text for every cp"""
        devc.setfont(font)
        devc.drawrotatedtext(text, pos_x, pos_y, 90)

    @staticmethod
    def gradtext(devc, text, pos_x, pos_y, font):
        """Draw gradient text for every cp"""
        devc.setfont(font)
        devc.drawtext(text, pos_x, pos_y)


class ThemeBclassic(Theme):
    """Bezier classic CycloGraph theme"""
    themename = "Smooth classic"
    shear = -0.15

    @staticmethod
    def getcolor(colors, levels, grad):
        """Return gradient relative color"""
        gradient = abs(grad)
        for i in range(len(levels)):
            if gradient < levels[i]:
                return colors[i-1]
        return colors[i]

    @staticmethod
    def drawslopeinfo(devc, info, settings, startx, starty):
        """Draw slope's name and other info background"""
        ThemeClassic.drawslopeinfo(devc, info, settings, startx, starty)

    @staticmethod
    def drawlegend(devc, settings, startx, starty, title):
        """Draw slope's legend"""
        ThemeClassic.drawlegend(devc, settings, startx, starty, title)

    @staticmethod
    def paintbackground(devc, width, height):
        """Draw backgroundr"""
        ThemeClassic.paintbackground(devc, width, height)

    @staticmethod
    def gradback(devc, width, height, setting_f, startcol="rgb(64, 185, 255)", endcol='white'):
        """Draw a gradient background"""
        ThemeClassic.gradback(devc, width, height, setting_f, startcol, endcol)

    @staticmethod
    def olines(devc, slope, margins, width, height, metersize):
        """Draw altitude line"""
        ThemeClassic.olines(devc, slope, margins, width, height, metersize)

    @staticmethod
    def alttext(devc, slope, margins, max_y, metersize):
        """draw altitude text aside"""
        ThemeClassic.alttext(devc, slope, margins, max_y, metersize)

    @staticmethod
    def yaxis(devc, slope, margins, max_y, metersize):
        """Draw y axis"""
        ThemeClassic.yaxis(devc, slope, margins, max_y, metersize)

    @staticmethod
    def xaxis(devc, slope, margins, max_y, d_incr, kmsize, dx, dy):
        """Draw km bar"""
        ThemeClassic.xaxis(devc, slope, margins, max_y, d_incr, kmsize, dx, dy)

    @staticmethod
    def fillhslopecontour(devc, spath_back, dx, dy, colorlist):
        """ Draw horizontal contour."""
        spath_back.reverse()
        index = 0
        for pnts in spath_back:
            devc.setbrush(colorlist[index])
            index = index + 1
            path = devc.startpath(pnts[0])
            devc.drawpathcubicto(path, pnts[1:])
            point = (pnts[3][0]+dx, pnts[3][1]+dy)
            points = [(pnts[2][0]+dx, pnts[2][1]+dy),
                    (pnts[1][0]+dx, pnts[1][1]+dy),
                    (pnts[0][0]+dx, pnts[0][1]+dy)]
            devc.drawpathlineto(path, point)
            pnts = list(pnts)
            pnts.reverse()
            devc.drawpathcubicto(path, points)
            devc.endpath(path)

    @staticmethod
    def fillfirsthpoly(devc, points, color):
        """fill first horizontal polygon of the 3d slope"""
        ThemeClassic.fillfirsthpoly(devc, points, color)

    @staticmethod
    def fillvslopecontour(devc, path_points, pixel_0m, pixel_1000m, vpolygons, colorlist):
        """Draw a close smooth path thant contais the whole slope."""
        index = 0
        start_pnt = path_points.pop(0)
        point = path_points[0]
        start_pnt = point[0]
        for pnt in path_points[:-1]:
            devc.setbrush(colorlist[index])
            path = devc.startpath(start_pnt)
            devc.drawpathcubicto(path, pnt[1:])
            devc.drawpathlineto(path, vpolygons[index][1])
            devc.drawpathlineto(path, vpolygons[index][0])
            devc.drawpathlineto(path, vpolygons[index][3])
            devc.endpath(path)
            start_pnt = pnt[3]
            index = index + 1

    @staticmethod
    def desctext(devc, text, pos_x, pos_y, font):
        """Draw description text for every cp"""
        ThemeClassic.desctext(devc, text, pos_x, pos_y, font)

    @staticmethod
    def gradtext(devc, text, pos_x, pos_y, font):
        """Draw gradient text for every cp"""
        ThemeClassic.gradtext(devc, text, pos_x, pos_y, font)


class ThemeBwhite(Theme):
    """bezier white theme"""
    themename = "giro"
    shear = -0.15

    @staticmethod
    def getcolor(colors, levels, grad):
        """Return gradient relative color"""
        ThemeClassic.getcolor(colors, levels, grad)

    @staticmethod
    def drawslopeinfo(devc, info, settings, startx, starty):
        """Draw slope's name and other info background"""
        ThemeClassic.drawslopeinfo(devc, info, settings, startx, starty)

    @staticmethod
    def paintbackground(devc, width, height):
        """Draw background"""
        ThemeClassic.paintbackground(devc, width, height)

    @staticmethod
    def gradback(devc, width, height, setting_f, startcol="rgb(0, 0, 0)", endcol="rgb(70, 80, 70)"):
        """Draw a gradient background"""
        devc.setfont(setting_f)  # FIXME: move setfont in a new method

    @staticmethod
    def alttext(devc, slope, margins, max_y, metersize):
        """draw altitude text aside"""
        ThemeClassic.alttext(devc, slope, margins, max_y, metersize)

    @staticmethod
    def yaxis(devc, slope, margins, max_y, metersize):
        """Draw y axis"""
        ThemeClassic.yaxis(devc, slope, margins, max_y, metersize)

    @staticmethod
    def xaxis(devc, slope, margins, max_y, d_incr, kmsize, dx, dy):
        """Draw km bar"""
        ThemeClassic.xaxis(devc, slope, margins, max_y, d_incr, kmsize, dx, dy)

    @staticmethod
    def fillhslopecontour(devc, spath_back, dx, dy, colorlist):
        """ Draw horizontal contour."""
        devc.setbrush("rgb(207, 208, 212)")
        spath_back.reverse()
        for pnts in spath_back:
            path = devc.startpath(pnts[0])
            devc.drawpathcubicto(path, pnts[1:])
            point = (pnts[3][0]+dx, pnts[3][1]+dy)
            points = [(pnts[2][0]+dx, pnts[2][1]+dy),
                    (pnts[1][0]+dx, pnts[1][1]+dy),
                    (pnts[0][0]+dx, pnts[0][1]+dy)]
            devc.drawpathlineto(path, point)
            pnts = list(pnts)
            pnts.reverse()
            devc.drawpathcubicto(path, points)
            devc.endpath(path)

    @staticmethod
    def fillfirsthpoly(devc, points, color):
        """fill first horizontal polygon of the 3d slope"""
        devc.setbrush("rgb(207, 208, 212)")
        devc.drawpolygon(points)

    @staticmethod
    def fillvslopecontour(devc, path_points, pixel_0m, pixel_1000m, vpolygons, colorlist):
        """Draw a close smooth path thant contais the whole slope."""
        devc.setlineargradientbrush([("rgb(255, 253, 241)", 0),
                                     ("rgb(255, 253, 241)", 0.25),
                                     ("rgb(222, 234, 190)", 0.50),
                                     ("rgb(240, 236, 189)", 0.75),
                                     ("rgb(233, 198, 153)", 1)],
                                    (0, pixel_0m), (0, pixel_0m + (pixel_1000m - pixel_0m)*1.8))
        start_pnt = path_points.pop(0)
        path = devc.startpath(start_pnt)
        ctrl_pnts = path_points.pop(0)
        devc.drawpathlineto(path, ctrl_pnts[0])
        devc.drawpathcubicto(path, ctrl_pnts[1:])
        for pnt in path_points[:-1]:
            devc.drawpathcubicto(path, pnt[1:])
        devc.drawpathlineto(path, path_points[-1])
        devc.endpath(path)

    @staticmethod
    def desctext(devc, text, pos_x, pos_y, font):
        """Draw description text for every cp"""
        ThemeClassic.desctext(devc, text, pos_x, pos_y, font)


class ThemeBlue(Theme):
    """Blue theme"""
    themename = "Blue"
    shear = -0.15

    @staticmethod
    def getcolor(colors, levels, grad):
        """Return gradient relative color"""
        gradient = abs(grad)
        for i in range(len(levels)):
            if gradient < levels[i]:
                return colors[i-1]
        return colors[i]

    @staticmethod
    def drawslopeinfo(devc, info, settings, startx, starty):
        """Draw slope's name and other info background"""
        name, avp, maxp, eleg, heightg = info
        devc.setfont(settings['ftitle'])
        devc.setpen('black', 1)
        devc.drawtext(name, startx, starty)
        if settings['sinfo']:
            devc.setfont(settings['fdesc'])
            devc.drawtext(avp, startx+10, starty+20)
            devc.drawtext(maxp, startx+10, starty+40)
            devc.drawtext(eleg, startx+10, starty+60)
            devc.drawtext(heightg, startx+10, starty+80)

    @staticmethod
    def paintbackground(devc, width, height):
        """Draw background"""
        devc.setbrush("rgb(255, 255, 255)")
        rec = (0, 0, width, height)
        devc.drawrectangle(*rec)

    @staticmethod
    def gradback(devc, width, height, setting_f, startcol="rgb(255, 255, 255)", endcol="rgb(250, 250, 250)"):
        """Draw a gradient background"""
        area = (0, 0, width, height)
        devc.gradfill(area, startcol, endcol)
        devc.setpen('black', 1)
        devc.setfont(setting_f)

    @staticmethod
    def olines(devc, slope, margins, width, height, metersize):
        """Draw altitude line"""
        devc.setpen("rgb(200, 200, 250)", 1)
        (updownmar, leftrightmar) = margins
        (lef_mar, rig_mar) = leftrightmar
        (upp_mar, low_mar) = updownmar
        x0_alt_bar = lef_mar - 3
        x1_alt_bar = width - rig_mar - 3
        y_alt_base_bar = height - low_mar
        for i in range(int(slope.min_h / slope.h_incr),
                       int((slope.max_h / slope.h_incr) + 1)):
            yincr = int((i * slope.h_incr - slope.min_h) * metersize)
            devc.drawline(x0_alt_bar, y_alt_base_bar - yincr,
                          x1_alt_bar, y_alt_base_bar - yincr)

    @staticmethod
    def alttext(devc, slope, margins, max_y, metersize):
        """draw altitude text aside"""
        (updownmar, leftrightmar) = margins
        (lef_mar, rig_mar) = leftrightmar
        (upp_mar, low_mar) = updownmar

        devc.setpen('black', 2)
        x_alt_txt = lef_mar - 40
        y_alt_base_txt = max_y - low_mar - 8
        for i in range(int(slope.min_h/slope.h_incr), int((slope.max_h/slope.h_incr) + 1)):
            yincr = int((i * slope.h_incr - slope.min_h) * metersize)
            alti_t = ("%4d" % (i*slope.h_incr))
            devc.drawtext(alti_t, x_alt_txt, y_alt_base_txt - yincr)

    @staticmethod
    def yaxis(devc, slope, margins, max_y, metersize):
        """Draw y axis"""
        ThemeClassic.yaxis(devc, slope, margins, max_y, metersize)

    @staticmethod
    def xaxis(devc, slope, margins, max_y, d_incr, kmsize, dx, dy):
        """Draw km bar"""
        devc.setpen('black', 1)
        ThemeClassic.xaxis(devc, slope, margins, max_y, d_incr, kmsize, dx, dy)

    @staticmethod
    def fillhpoly(devc, points, color):
        """fill horizontal polygons of the 3d slope"""
        devc.setpen('black', 1)
        devc.setbrush("rgb(220, 220, 250)")
        devc.drawpolygon(points)

    @staticmethod
    def fillfirsthpoly(devc, points, color):
        """fill first horizontal polygon of the 3d slope"""
        devc.setpen('black', 1)
        # devc.setbrush(color)
        devc.setlineargradientbrush([("rgb(250, 250, 250)", 0),
                                     ("rgb(200, 200, 250)", 1)],
                                    (0, points[3][1]), (0, devc.size_y/2))
        devc.drawpolygon(points)

    @staticmethod
    def fillvpoly(devc, points, color):
        """fill vertical polygons (2d and 3d) of the slope"""
        devc.setpen('black', 1)
        # devc.setbrush(color)
        devc.setlineargradientbrush([("rgb(250, 250, 250)", 0),
                                     ("rgb(200, 200, 250)", 1)],
                                    (0, points[1][1]), (0, devc.size_y/2))  # max(points[3][1], points[2][1])))
        devc.drawpolygon(points)

    @staticmethod
    def desctext(devc, text, pos_x, pos_y, font):
        """Draw description text for every cp"""
        devc.setpen('black', 1)
        devc.setfont(font)
        devc.drawrotatedtext(text, pos_x, pos_y, 90)

    @staticmethod
    def gradtext(devc, text, pos_x, pos_y, font):
        """Draw gradient text for every cp"""
        devc.setpen('black', 1)
        devc.setfont(font)
        devc.drawtext(text, pos_x, pos_y)


class ThemeBlack(Theme):
    """Black theme"""
    themename = "Black"
    shear = -0.15

    @staticmethod
    def getcolor(colors, levels, grad):
        """Return gradient relative color"""
        gradient = abs(grad)
        for i in range(len(levels)):
            if gradient < levels[i]:
                return colors[i-1]
        return colors[i]

    @staticmethod
    def drawslopeinfo(devc, info, settings, startx, starty):
        """Draw slope's name and other info background"""
        name, avp, maxp, eleg, heightg = info
        devc.setfont(settings['ftitle'])
        devc.setpen('white', 1)
        devc.drawtext(name, startx, starty)
        if settings['sinfo']:
            devc.setfont(settings['fdesc'])
            devc.drawtext(avp, startx+10, starty+20)
            devc.drawtext(maxp, startx+10, starty+40)
            devc.drawtext(eleg, startx+10, starty+60)
            devc.drawtext(heightg, startx+10, starty+80)

    @staticmethod
    def drawlegend(devc, settings, startx, starty, title):
        """Draw slope's legend"""
        devc.setbrush("rgb(90, 100, 90)")
        rec = (startx, starty, 310, 52)
        devc.setpen('black', 1)
        devc.drawrectangle(*rec)
        devc.setfont(settings['fdesc'])
        devc.setpen('white', 2)
        devc.drawtext(title, startx+130, starty+2)
        incr = 43
        for i in range(len(settings['colors'])):
            devc.setbrush(settings['colors'][i])
            devc.setpen('black', 1)
            rec = (startx+8+i*incr, starty+20, 36, 10)
            devc.drawrectangle(*rec)
            devc.setpen('white', 1)
            alti_t = (">%d%%" % (settings['levels'][i]))
            devc.drawtext(alti_t, startx+9+i*incr, starty+34)

    @staticmethod
    def paintbackground(devc, width, height):
        """Draw background"""
        devc.setbrush("rgb(70, 80, 70)")
        rec = (0, 0, width, height)
        devc.drawrectangle(*rec)

    @staticmethod
    def gradback(devc, width, height, setting_f, startcol="rgb(0, 0, 0)", endcol="rgb(70, 80, 70)"):
        """Draw a gradient background"""
        area = (0, 0, width, height)
        devc.gradfill(area, startcol, endcol)
        devc.setpen('black', 1)
        devc.setfont(setting_f)

    @staticmethod
    def alttext(devc, slope, margins, max_y, metersize):
        """draw altitude text aside"""
        (updownmar, leftrightmar) = margins
        (lef_mar, rig_mar) = leftrightmar
        (upp_mar, low_mar) = updownmar

        devc.setpen('white', 2)
        x_alt_txt = lef_mar - 40
        y_alt_base_txt = max_y - low_mar - 8
        for i in range(int(slope.min_h/slope.h_incr), int((slope.max_h/slope.h_incr) + 1)):
            yincr = int((i * slope.h_incr - slope.min_h) * metersize)
            alti_t = ("%4d" % (i*slope.h_incr))
            devc.drawtext(alti_t, x_alt_txt, y_alt_base_txt - yincr)

    @staticmethod
    def yaxis(devc, slope, margins, max_y, metersize):
        """Draw y axis"""
        ThemeClassic.yaxis(devc, slope, margins, max_y, metersize)

    @staticmethod
    def xaxis(devc, slope, margins, max_y, d_incr, kmsize, dx, dy):
        """Draw km bar"""
        (updownmar, leftrightmar) = margins
        (lef_mar, rig_mar) = leftrightmar
        (upp_mar, low_mar) = updownmar
        for i in range(int(int(slope.max_d) / d_incr - int(slope.min_d) / d_incr + 1)):
            if i % 2 == 0:
                devc.setbrush("rgb(0, 0, 0)")
            else:
                devc.setbrush("rgb(255, 255, 255)")
            rec = (i * d_incr * kmsize + lef_mar + dx,
                   max_y - low_mar + dy, d_incr * kmsize + 1, 5)
            devc.setpen('black', 1)
            devc.drawrectangle(*rec)
            devc.setpen('white', 1)
            d_label = ("%d" % (i * d_incr + int(slope.min_d)))
            devc.drawtext(d_label, rec[0], rec[1] + 5)

    @staticmethod
    def fillhpoly(devc, points, color):
        """fill horizontal polygons of the 3d slope"""
        devc.setpen('black', 1)
        devc.setbrush(color)
        devc.drawpolygon(points)

    @staticmethod
    def fillfirsthpoly(devc, points, color):
        """fill first horizontal polygon of the 3d slope"""
        devc.setpen('black', 1)
        devc.setbrush(color)
        devc.setlineargradientbrush([("rgb(0, 0, 0)", 0),
                                     (color, 1)],
                                    (0, points[3][1]), (0, points[1][1]))
        devc.drawpolygon(points)

    @staticmethod
    def fillvpoly(devc, points, color):
        """fill vertical polygons (2d and 3d) of the slope"""
        devc.setpen('black', 1)
        devc.setbrush(color)
        devc.setlineargradientbrush([("rgb(0, 0, 0)", 0),
                                     (color, 1)],
                                    (0, points[1][1]), (0, min(points[3][1], points[2][1])))
        devc.drawpolygon(points)

    @staticmethod
    def desctext(devc, text, pos_x, pos_y, font):
        """Draw description text for every cp"""
        devc.setpen('white', 1)
        devc.setfont(font)
        devc.drawrotatedtext(text, pos_x, pos_y, 90)

    @staticmethod
    def gradtext(devc, text, pos_x, pos_y, font):
        """Draw gradient text for every cp"""
        devc.setpen('white', 1)
        devc.setfont(font)
        devc.drawtext(text, pos_x, pos_y)


class ThemeGray(Theme):
    """Gray theme"""
    themename = "Gray"
    shear = -0.15

    @staticmethod
    def getcolor(colors, levels, grad):
        """Return gradient relative color"""
        gradient = abs(grad)
        graycolors = ["rgb(175,175,175)", "rgb(150,150,150)",
                      "rgb(125,125,125)", "rgb(100,100,100)",
                      "rgb(75,75,75)", "rgb(50,50,50)", "rgb(25,25,25)"]
        for i in range(len(levels)):
            if gradient < levels[i]:
                return graycolors[i-1]
        return graycolors[i]

    @staticmethod
    def drawslopeinfo(devc, info, settings, startx, starty):
        """Draw slope's name and other info background"""
        ThemeClassic.drawslopeinfo(devc, info, settings, startx, starty)

    @staticmethod
    def drawlegend(devc, settings, startx, starty, title):
        """Draw slope's legend"""
        graycolors = ["rgb(175,175,175)", "rgb(150,150,150)",
                      "rgb(125,125,125)", "rgb(100,100,100)",
                      "rgb(75,75,75)", "rgb(50,50,50)", "rgb(25,25,25)"]
        devc.setbrush("rgb(236, 236, 236)")
        rec = (startx, starty, 310, 52)
        devc.drawrectangle(*rec)
        devc.setfont(settings['fdesc'])
        devc.setpen('black', 2)
        devc.drawtext(title, startx+130, starty+2)
        incr = 43
        devc.setpen('black', 1)
        for i in range(len(graycolors)):
            devc.setbrush(graycolors[i])
            rec = (startx+8+i*incr, starty+20, 36, 10)
            devc.drawrectangle(*rec)
            alti_t = (">%d%%" % (settings['levels'][i]))
            devc.drawtext(alti_t, startx+9+i*incr, starty+34)

    @staticmethod
    def paintbackground(devc, width, height):
        """Draw background"""
        devc.setbrush('white')
        rec = (0, 0, width, height)
        devc.drawrectangle(*rec)

    @staticmethod
    def gradback(devc, width, height, setting_f, startcol="rgb(0, 0, 0)", endcol="rgb(70, 80, 70)"):
        """Draw a gradient background"""
        devc.setpen('black', 1)
        devc.setfont(setting_f)

    @staticmethod
    def olines(devc, slope, margins, width, height, metersize):
        """Draw altitude line"""
        ThemeClassic.olines(devc, slope, margins, width, height, metersize)

    @staticmethod
    def alttext(devc, slope, margins, max_y, metersize):
        """draw altitude text aside"""
        (updownmar, leftrightmar) = margins
        (lef_mar, rig_mar) = leftrightmar
        (upp_mar, low_mar) = updownmar

        devc.setpen('black', 2)
        x_alt_txt = lef_mar - 40
        y_alt_base_txt = max_y - low_mar - 8
        for i in range(int(slope.min_h/slope.h_incr), int((slope.max_h/slope.h_incr) + 1)):
            yincr = int((i * slope.h_incr - slope.min_h) * metersize)
            alti_t = ("%4d" % (i*slope.h_incr))
            devc.drawtext(alti_t, x_alt_txt, y_alt_base_txt - yincr)

    @staticmethod
    def yaxis(devc, slope, margins, max_y, metersize):
        """Draw y axis"""
        ThemeClassic.yaxis(devc, slope, margins, max_y, metersize)

    @staticmethod
    def xaxis(devc, slope, margins, max_y, d_incr, kmsize, dx, dy):
        """Draw km bar"""
        devc.setpen('black', 1)
        ThemeClassic.xaxis(devc, slope, margins, max_y, d_incr, kmsize, dx, dy)

    @staticmethod
    def fillhpoly(devc, points, color):
        """fill horizontal polygons of the 3d slope"""
        devc.setpen('black', 1)
        devc.setbrush(color)
        devc.drawpolygon(points)

    @staticmethod
    def fillfirsthpoly(devc, points, color):
        """fill first horizontal polygon of the 3d slope"""
        devc.setpen('black', 1)
        devc.setbrush(color)
        devc.setlineargradientbrush([("rgb(0, 0, 0)", 0),
                                     (color, 1)],
                                    (0, points[3][1]), (0, points[1][1]))
        devc.drawpolygon(points)

    @staticmethod
    def fillvpoly(devc, points, color):
        """fill vertical polygons (2d and 3d) of the slope"""
        devc.setpen('black', 1)
        devc.setbrush("rgb(100, 100, 100)")
        devc.setbrush(color)
        devc.setlineargradientbrush([("rgb(0, 0, 0)", 0),
                                     (color, 1)],
                                    (0, points[1][1]), (0, min(points[3][1], points[2][1])))
        devc.drawpolygon(points)

    @staticmethod
    def desctext(devc, text, pos_x, pos_y, font):
        """Draw description text for every cp"""
        devc.setfont(font)
        devc.setpen('white', 1)
        devc.drawrotatedtext(text, pos_x + 1, pos_y + 1, 90)
        devc.setpen('black', 1)
        devc.drawrotatedtext(text, pos_x, pos_y, 90)

    @staticmethod
    def gradtext(devc, text, pos_x, pos_y, font):
        """Draw gradient text for every cp"""
        devc.setfont(font)
        devc.setpen('black', 1)
        devc.drawtext(text, pos_x + 1, pos_y + 1)
        devc.setpen('white', 1)
        devc.drawtext(text, pos_x, pos_y)

# vim:sw=4:softtabstop=4:expandtab



# -*- coding: utf-8 -*-
# glal.py
"""This module provides a graphic library abstraction layer for Cyclograph"""

# Copyright (C) 2008-2015 Federico Brega, Pierluigi Villani

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

from __future__ import print_function, unicode_literals

import sys
import glal_selected
LIBRARY = glal_selected.library
(FD_OPEN, FD_SAVE) = range(2)


def Message(*args):
    """This funcion return the class that wraps the using messaging library"""
    if LIBRARY == 'cli':
        return Message_cli(*args)


def DeviceContext(*args):
    if LIBRARY == 'cli':
        return DeviceContext_cli(*args)


def Image(*args):
    if LIBRARY == 'cli':
        return Image_cli(*args)


def addstatusbartext(maingui, text):
    """Add text to main gui status bar"""
    if LIBRARY == 'cli':
        sys.stderr.write(text)


def FileDialog(parent, title, dirname, other, file_types, rw, filename=None):
    """Return File dialog """
    return gui.FileDialog(parent, title, dirname, filename, other,
                               file_types, rw, FD_OPEN, FD_SAVE)

# Begin of cli classes ###


class Message_cli():
    """CLI interfaces doesn't need to update their view
    because they haven't got any
    """
    def send(self, *args):
        """ send method not used for cli"""
        pass

    def subscribe(self, *args):
        """ subscribe method not used for cli"""
        pass


def DeviceContext_cli(*args):
    """ Get a device contex for CLI default is an svg generator."""
    return DeviceContext_svg(*args)


class Image_cli():
    """ Image cli not used"""
    pass
# End of cli classes ###


# Begin of SVG ###
class DeviceContext_svg():
    """Device Context class to draw sgv files"""
    colors = {'white': '#ffffff',
              'black': '#000000'}

    def setcolor(self, color):
        """ Set color converting to svg compatible format."""
        if type(color) == type(str()):
            if color.startswith('#'):
                return color
            elif color.startswith('rgb'):
                tuple_ = [int(x) for x in color[4:-1].split(',')]
                return tupletocss(tuple_)
            else:
                return self.colors[color]
        else:
            sys.stderr.write("Color type unknown: "+color+" "+str(type(color))+"\n")
            return False

    def init_draw_surf(self, size, p3d):
        """ Init draw surface"""
        self.size_x = size[0]
        self.size_y = size[1]
        self.pencolor = None
        self.lineargcount = 0
        self.pensize = 0
        self.pshear = p3d
        self.svg = '''<?xml version="1.0" standalone="no"?>
<!-- Created with CycloGraph (http://cyclograph.sourceforge.net) -->
<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN"
"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">
<svg width="%d" height="%d" version="1.1"
xmlns="http://www.w3.org/2000/svg">\n
''' % size
        if p3d:
            self.brushcolor = '#ffffff'  # white background
            self.svg += '''<rect x="0" y="0" width="%i" height="%i" fill=" %s"/>
''' % (self.size_x, self.size_y, self.brushcolor)

    def shear(self, shear):
        if shear != 0:
            self.pshear = True
            self.svg += '<g transform="skewY(%f)">' % (shear*50,)
        else:
            self.pshear = False

    def getsize(self):
        """ Return size"""
        return (self.size_x, self.size_y)

    def gradfill(self, rect, startcolor, endcolor):
        """ Fill gradient"""
        self.svg += '''
<defs>
<linearGradient id="background" x1="0%%" y1="0%%" x2="0%%" y2="100%%">
<stop offset="0%%" style="stop-color:%s;stop-opacity:1;"/>
<stop offset="80%%" style="stop-color:%s;stop-opacity:1;"/>
</linearGradient>
</defs>
<rect width="%d" height="%d" style="fill:url(#background)"/>
''' % (self.setcolor(startcolor), self.setcolor(endcolor),
        rect[2] - rect[0], rect[3] - rect[1])

    def setpen(self, color, size):
        """ Set pen's size and color"""
        self.pencolor = self.setcolor(color)
        self.pensize = str(size)

    def setfont(self, fontdict):
        """ Set font"""
        self.dim = fontdict['dim']
        self.font = "font-family:%s;font-size:%spx" % (fontdict['des'], self.dim)

    def drawtext(self, text, pos_x, pos_y):
        """ Draw text at position pos_x,pos_y """
        self.drawrotatedtext(text, pos_x, pos_y, 0)

    def gettextwidth(self, text):
        """ Return text length"""
        return self.dim*len(text)-30  # TODO: check if there is a more accurate method

    def gettextheight(self, text):
        """ Return text height"""
        return self.dim  # TODO: check if there is a more accurate method

    def drawline(self, pos_x0, pos_y0, pos_x1, pos_y1):
        """ draw line"""
        self.svg += '''<line x1="%d" y1="%d" x2="%d" y2="%d"
style="stroke:%s;stroke-width:%s"/>
''' % (pos_x0, pos_y0, pos_x1, pos_y1, self.pencolor, self.pensize)

    def setlineargradientbrush(self, colorlist, startp, endp):
        """ Get a linear gradient from startp to endp, using colors in colorlist.
        The elments of colorlist are tuple in the format (color, realtive position)."""
        self.svg += '''<defs>
<linearGradient id="linearg%s" gradientUnits="userSpaceOnUse" x1="%s" y1="%s" x2="%s" y2="%s">
''' % (self.lineargcount, startp[0], startp[1], endp[0], endp[1])
        for color in colorlist:
            self.svg += '''<stop offset="%s%%" style="stop-color:%s;stop-opacity:1;"/>
''' % (color[1]*100, self.setcolor(color[0]))
        self.svg += '''</linearGradient>
</defs>
'''
        self.brushcolor = "url(#linearg%s)" % (self.lineargcount)
        self.lineargcount = self.lineargcount + 1

    def setbrush(self, color):
        """ Set brush color"""
        self.brushcolor = self.setcolor(color)

    def drawrectangle(self, pos_x0, pos_y0, width, height):
        """ Draw rectangle"""
        self.svg += '''<rect x="%i" y="%i" width="%i" height="%i"
style="fill: %s;stroke:%s;stroke-width:%s"/>
''' % (pos_x0, pos_y0, width, height, self.brushcolor,
       self.pencolor, self.pensize)

    def drawrotatedtext(self, text, pos_x, pos_y, angle):
        """Draw the text given, rotated by the angle."""
        dy = self.dim
        self.svg += '<text x="%i" y="%i" style="%s;fill:%s"' % (pos_x, pos_y+dy, self.font, self.pencolor)
        if angle:
            self.svg += ' transform="rotate(%d,%d,%d)"' % (-angle, pos_x, pos_y)
        self.svg += '''>%s</text>
''' % (text)

    def drawpolygon(self, sequence):
        """ Draw polygon"""
        if len(sequence) < 2:
            return
        self.svg += '<polygon points="'
        for pt in sequence:
            self.svg += '%d, %d ' % (pt[0], pt[1])
        self.svg += '"\n style="fill:%s;stroke:%s;stroke-width:%s"/>\n' \
                % (self.brushcolor, self.pencolor, self.pensize)

    def startpath(self, point):
        """ Start a path in the specified point,"""
        self.svg += '''<path style="fill:%s;stroke:%s;stroke-width:%s;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1"
d="M %d,%d''' % (self.brushcolor, self.pencolor, self.pensize, point[0], point[1])
        self.path = point
        return self.path

    def drawpathlineto(self, path, point):
        """ Draw a straight line from the last point to the given point."""
        self.svg += ''' L %d,%d''' % (point[0], point[1])
        self.path = point

    def drawpathcubicto(self, path, controlpoints):
        """ Draw a cubic Beziér frome the last point using the given list of
        three control points."""
        self.svg += ''' C %d,%d %d,%d %d,%d''' % (controlpoints[0][0], controlpoints[0][1],
        controlpoints[1][0], controlpoints[1][1],
        controlpoints[2][0], controlpoints[2][1])
        self.path = controlpoints[2]

    def endpath(self, path):
        """ Show the path."""
        self.svg += '"/>\n'

    def end_draw(self):
        """ end drawing"""
        if self.pshear:
            self.svg += '</g>'
        self.svg += "</svg>\n"
        print(self.svg)


class Image_svg():
    """ svg image"""
    def __init__(self, size_x,  size_y, plotfnct):
        self.size = (size_x,  size_y)
        self.plotfnct = plotfnct
        self.svg = ""

    def plot(self, settings):
        """draw svg image"""
        dcwrpp = DeviceContext_svg()
        dcwrpp.init_draw_surf(self.size, settings['3d'])
        self.plotfnct(settings, dcwrpp)
        # dcwrpp.end_draw()
        self.svg = dcwrpp.svg
        if settings['3d']:
            self.svg += "</g>\n"
        self.svg += "</svg>\n"

    def savetofile(self, path, format):
        """ Save to file"""
        if format == 'svg':
            try:
                fid = open(path, 'w')
                fid.write(self.svg)
            finally:
                fid.close()


def tupletocss(tuple_):
    """ convert tuple to css"""
    numb = tuple_[0] << 16 | tuple_[1] << 8 | tuple_[2]
    return '#' + hex(numb).split('x')[1].zfill(6)
# End of SVG ###

if LIBRARY == 'qt':
    from qt.glal import *

elif LIBRARY == 'gtk3':
    from gtk3.glal import *

elif LIBRARY == 'cli':
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# gtk3.glal.py
"""This module provides a graphic library abstraction layer for Cyclograph"""

# Copyright (C) 2008-2016 Federico Brega, Pierluigi Villani

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import math
import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GObject, Gdk, GLib
import cairo
from . import gui

GObject.threads_init()


def initapp():
    return Gtk.main()


def bind(cgfunc, menuopt):
    """Connect gui menu option with cg function"""
    return menuopt.connect("activate", cgfunc)


def bind_close(cg_exit, gui):
    """Connect gui close event with cg exit function"""
    gui.cg_exit = cg_exit
    bind(gui.exit, gui.menu_item_exit)


def ToolbarBind(fun, button):
    """Connect gui toolbar button with cg function"""
    return button.connect("clicked", fun)


def OptionCheck(guiopt, option):
    """Set a gui menu option to selected/unselected """
    guiopt.set_active(option)


def enable_saving(gui, bool_val):
    """Enable or disable saving options according to bool_val"""
    gui.menu_item_save.set_sensitive(bool_val)
    gui.menu_item_save_as.set_sensitive(bool_val)
    gui.action_add.set_sensitive(bool_val)
    gui.action_edit.set_sensitive(bool_val)
    gui.action_delete.set_sensitive(bool_val)
    gui.action_plot.set_sensitive(bool_val)
    gui.action_map.set_sensitive(bool_val)
    gui.action_properties.set_sensitive(bool_val)


def addstatusbartext(maingui, text):
    """Add text to main gui status bar"""
    maingui.addstatusbartext(text)
    maingui.sbtimer.start()


def signalbug(event=None):
    """Open webpage to signal bugs"""
    bugUrl = 'http://sourceforge.net/project/memberlist.php?group_id=227295'
    # Looks like gtk hasn't any function to launch an external browser,
    # the best we can do is use python webbrowser
    import webbrowser
    webbrowser.open(bugUrl, new=True, autoraise=True)


class Message():
    """ Gtk3 message"""
    def __init__(self):
        class Messenger(GObject.GObject):
            __gsignals__ = {
                str("UPDATE_TAB"): (GObject.SignalFlags.ACTION, None,
                                    (int, int, )),
                str("SLOPE_ADD"): (GObject.SignalFlags.ACTION, None,
                                   (int, int, )),
                str("SLOPE_DEL"): (GObject.SignalFlags.ACTION, None,
                                   (int, int, )),
                }
        self.gobj = Messenger()

    def send(self, message, slope_number, row_num):
        """ Send message"""
        self.gobj.emit(message, slope_number, row_num)

    def subscribe(self, function, emitter, message):
        """ Subscribe message"""
        self.handler = function
        emitter.gobj.connect(message, self._reformat_handler_param)

    def _reformat_handler_param(self, emitter, arg1, arg2):
        self.handler(arg1, arg2)


class Notebook:
    """Gtk3 notebook"""
    def __init__(self, notebook, fun_close):
        self.notebook = notebook
        self.fun_close = fun_close
        self.notebook.set_scrollable(True)

    def Page(self):
        """Creates page"""
        return gui.Page(self.notebook)

    def set_page_label(self, page_num,  text):
        """Set page label text"""
        lab = self.notebook.get_tab_label(self.get_page(page_num))
        lab.label.set_text(text)

    def add_page(self, page, title):
        """Add page to the notebook"""
        box = gui.tab_label(page, title)
        page.button_close = box.closebtn
        page.button_close.connect("clicked", self.fun_close, page)
        self.notebook.append_page(page, box)
        self.notebook.show_all()
        nbpages = self.notebook.get_n_pages()
        self.notebook.set_current_page(nbpages-1)

    def remove_page(self, page_num):
        """Remove page from the notebook"""
        self.notebook.remove_page(page_num)

    def getselpagnum(self):
        """Return current selected page"""
        return self.notebook.get_current_page()

    def setselpagnum(self, page_num):
        """Set current selected page"""
        self.notebook.set_current_page(page_num)

    def get_page(self, page_num):
        """Return page at page_num"""
        return self.notebook.get_nth_page(page_num)

    def get_pagenum(self, argument):
        """Return page_num from page passed as argument"""
        return self.notebook.page_num(argument[1])


class DeviceContext():
    """Device Context Gtk3 class wrapper"""
    def _convcolor(self, color):
        """ Convert color"""
        col = color
        if isinstance(color, str):
            if color.startswith("rgb"):
                col = [(int(component) << 8)/65535.0 for component in color[4:-1].split(',')]
            elif color == "white":
                col = [1, 1, 1]
            else:
                col = [0, 0, 0]
        return col

    def init_draw_surf(self, widget, width, height):
        """Initialize device context"""
        self.size_x = width
        self.size_y = height
        self.devc = widget

        self.FONT_TYP = {
                         "light": cairo.FONT_WEIGHT_NORMAL,
                         "normal": cairo.FONT_WEIGHT_NORMAL,
                         "bold": cairo.FONT_WEIGHT_BOLD,
                        }
        # pen color
        self.pcolor = [0, 0, 0]
        # brush color
        self.bcolor = [0, 0, 0]

    def shear(self, shear):
        """ device shear"""
        matrix = cairo.Matrix(yx = shear)
        self.devc.set_matrix(matrix)

    def getsize(self):
        """Return device context dimensions"""
        return (self.size_x, self.size_y)

    def gradfill(self, rect, startcolor, endcolor):
        """Fill context with gradient"""
        (startx, starty) = (rect[0], rect[1])
        scolor = self._convcolor(startcolor)
        ecolor = self._convcolor(endcolor)
        lg1 = cairo.LinearGradient(0, 160.0,  0, self.size_y)
        lg1.add_color_stop_rgb(0, *scolor)
        lg1.add_color_stop_rgb(1, *ecolor)
        self.devc.rectangle(startx, starty, self.size_x, self.size_y)
        self.devc.set_source(lg1)
        self.devc.fill()

    def setpen(self, color, size):
        """Set pen color and size"""
        self.devc.set_line_width(size)
        self.pcolor = self._convcolor(color)

    def setfont(self, font):
        """Set font"""
        try:
            self.devc.select_font_face(font['des'], cairo.FONT_SLANT_NORMAL,
                                       self.FONT_TYP[font['typ']])
            self.devc.set_font_size(font['dim'])
        except Exception:
            pass

    def drawtext(self, text, pos_x, pos_y):
        """Draw text"""
        self.devc.set_source_rgb(*self.pcolor)
        self.devc.move_to(pos_x, pos_y+12)
        self.devc.show_text(text)

    def gettextwidth(self, text):
        """ Return text length"""
        return int(self.devc.text_extents(text)[2])

    def gettextheight(self, text):
        """ Return text height"""
        return int(self.devc.text_extents(text)[3])

    def drawline(self, pos_x0, pos_y0, pos_x1, pos_y1):
        """Draw line"""
        self.devc.set_source_rgb(*self.pcolor)
        self.devc.move_to(pos_x0, pos_y0)
        self.devc.line_to(pos_x1, pos_y1)
        self.devc.stroke()

    def setlineargradientbrush(self, colorlist, startp, endp):
        """ Get a linear gradient from startp to endp, using colors in colorlist.
        The elments of colorlist are tuple in the format (color, realtive position)."""
        grad = cairo.LinearGradient(startp[0], startp[1], endp[0], endp[1])
        for color in colorlist:
            grad.add_color_stop_rgb(color[1], *self._convcolor(color[0]))
        self.pattern = grad

    def setbrush(self, color):
        """Set brush color"""
        self.pattern = cairo.SolidPattern(*self._convcolor(color))

    def drawrectangle(self, pos_x0, pos_y0, width, height):
        """Draw rectangle"""
        self.devc.set_source_rgb(0, 0, 0)
        self.devc.rectangle(pos_x0, pos_y0, width, height)
        self.devc.stroke_preserve()
        self.devc.set_source(self.pattern)
        self.devc.fill()

    def drawrotatedtext(self, text, pos_x, pos_y, angle):
        """Draw rotated text"""
        self.devc.save()
        self.devc.translate(pos_x, pos_y)
        self.devc.rotate(-math.radians(angle))
        self.drawtext(text, 0, 0)
        self.devc.restore()

    def drawpolygon(self, sequence):
        """Draw polygon"""
        self.devc.set_source_rgb(0, 0, 0)
        self.devc.move_to(*sequence[0])
        for point in sequence:
            self.devc.line_to(*point)
        self.devc.close_path()
        self.devc.stroke_preserve()
        self.devc.set_source(self.pattern)
        self.devc.fill()

    def startpath(self, point):
        """ Start a path in the specified point,"""
        (pointx, pointy) = point
        self.devc.set_source_rgb(0, 0, 0)
        self.devc.move_to(pointx, pointy)
        return None

    def drawpathlineto(self, path, point):
        """ Draw a straight line from the last point to the given point."""
        (pointx, pointy) = point
        self.devc.line_to(pointx, pointy)

    def drawpathcubicto(self, path, controlpoints):
        """ Draw a cubic Beziér from the last point using the given list of
        three control points."""
        points = []
        for pnt in controlpoints:
            points.append(pnt[0])
            points.append(pnt[1])
        self.devc.curve_to(*points)

    def endpath(self, path):
        """ Show the path."""
        self.devc.set_source_rgb(0, 0, 0)
        self.devc.stroke_preserve()
        self.devc.set_source(self.pattern)
        self.devc.fill()

    def end_draw(self):
        """End drawing not used for gtk gui"""
        pass


class Image():
    """Gtk3 image class wrapper"""
    def __init__(self, size_x, size_y, plotfnct):
        (self.size_x, self.size_y) = (size_x,  size_y)
        self.plotfnct = plotfnct
        self.image = cairo.ImageSurface(cairo.FORMAT_RGB24,
                                         self.size_x, self.size_y)

    def plot(self, settings):
        """Plot image"""
        dcwrpp = DeviceContext()
        dcwrpp.init_draw_surf(cairo.Context(self.image),
                              self.size_x, self.size_y)
        self.plotfnct(settings, dcwrpp)

    def savetofile(self, path, format):
        """Save slope to image"""
        if format.lower() == 'png':
            self.image.write_to_png(path)
        if format.lower() == 'bmp':
            width, height = self.image.get_width(), self.image.get_height()
            pixbuf = Gdk.pixbuf_get_from_surface(self.image, 0, 0, width, height)
            pixbuf.savev(path, "bmp", [], [])
        if format.lower() == 'jpg':
            width, height = self.image.get_width(), self.image.get_height()
            pixbuf = Gdk.pixbuf_get_from_surface(self.image, 0, 0, width, height)
            pixbuf.savev(path, "jpeg", ["quality"], ["100"])


class Pdf():
    """Gtk3 pdf class wrapper"""
    def __init__(self, filepath):
        width, height = 793, 1122
        self.dy = 90
        self.dx = 300
        self.y_incr = 30
        self.surface = cairo.PDFSurface(filepath, width, height)
        self.context = cairo.Context(self.surface)
        # white background
        self.context.set_source_rgb(1, 1, 1)
        self.context.rectangle(0, 0, width, height)
        self.context.fill()
        self.context.select_font_face("Arial")
        self.context.set_font_size(18)

    def plot_image(self, settings, size_x, size_y, plotfnct):
        """Plot image"""
        image = cairo.SVGSurface(None, size_x, size_y)
        dcwrpp = DeviceContext()
        dcwrpp.init_draw_surf(cairo.Context(image), size_x, size_y)
        plotfnct(settings, dcwrpp)
        self.context.set_source_surface(image, 0, 0)
        self.context.paint()
        self.dy += size_y

    def addtext(self, text):
        self.dy += self.y_incr
        self.context.set_source_rgb(0, 0, 0)
        self.context.move_to(self.dx, self.dy)
        self.context.show_text(text)

    def addtitle(self, text):
        self.dy += self.y_incr
        self.context.set_source_rgb(0, 0, 0.4)
        self.context.move_to(self.dx-50, self.dy)
        self.context.show_text(text+":")

    def save(self):
        """Save slope to pdf"""
        self.context.show_page()
        self.surface.finish()


class ProgressDialog():
    """Gtk3 Progress dialog"""
    def __init__(self):
        self.pdialog = gui.ProgressDialog(_("Downloading altitudes"),
            _("Please wait, Cyclograph is downloading altitudes."), 1000.0)

    def update(self, value):
        """Update the progress shown and return if user want to abort."""
        return self.pdialog.update(value)

    def destroy(self):
        """Destroy dialog"""
        self.pdialog.destroy()


class Timer():
    """Gtk3 timer"""
    def __init__(self, period, callback):
        self.timer = 0
        self.callback = callback
        self.period = period

    def start(self):
        """Start timer"""
        self.timer = GLib.timeout_add(self.period, self.callback)

    def stop(self):
        """Stop timer"""
        GObject.source_remove(self.timer)
        self.timer = 0

# vim:sw=4:softtabstop=4:expandtab



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""cyclograph.py """

# Copyright (C) 2008-2017 Federico Brega, Pierluigi Villani

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import slope
import iofile
import glal
import altitude_downloader
import report_html
from themes import ThemeManager
import os
import gettext
import configparser
import sys
import queue
import argparse
from version import VERSION, LastVersionQuery

# font-typ light, normal, bold
DEFAULT_CONF = {"version": "1.2",
                "orizzontal lines": "True",
                "legend": "True",
                "slopeinfo": "True",
                "effect-3d": "True",
                "theme": "Classic",
                "font-des 1": "Sans",
                "font-typ 1": "normal",
                "font-dim 1": "10",
                "font-des g": "Sans",
                "font-typ g": "normal",
                "font-dim g": "10",
                "font-des t": "Sans",
                "font-typ t": "bold",
                "font-dim t": "20",
                "width": "800",
                "height": "600",
                "color 1": "rgb(0,0,255)",
                "color 2": "rgb(0,255,0)",
                "color 3": "rgb(255,255,0)",
                "color 4": "rgb(255,0,0)",
                "color 5": "rgb(200,0,0)",
                "color 6": "rgb(150,0,0)",
                "color 7": "rgb(100,0,0)",
                "level 1": "0",
                "level 2": "3.0",
                "level 3": "6.0",
                "level 4": "9.0",
                "level 5": "12.0",
                "level 6": "15.0",
                "level 7": "18.0",
                "server": "geonames.org",
                "numcps": "20",
                "widthsize": "540",
                "heightsize": "400",
                "maximized": "False",
                "latestVersionCheck": "True",
                "authorname": "",
                "authormail": "",
                }

SECTIONS = ["global", "font", "resolution", "colors", "levels", "kml", "main window"]


class Cyclograph():
    """Controller according to MCV"""
    def __init__(self):
        self.slopelist = slope.SlopeList()
        self.init_configparser()
        size = (self.config.getint("main window", "widthsize"),
                self.config.getint("main window", "heightsize"),
                self.config.getboolean("main window", "maximized"))
        self.mygui = glal.gui.Gui(None, "CycloGraph", size)
        self.notebook = glal.Notebook(self.mygui.notebook, self.close)
        self.slopecounter = 1

        self.setguiopt()

        glal.bind(self.new_slope, self.mygui.menu_item_new)
        glal.bind(self.open_slope, self.mygui.menu_item_open)
        glal.bind(self.create_kml, self.mygui.menu_item_createkml)
        glal.bind(self.save_slope, self.mygui.menu_item_save)
        glal.bind(self.save_slope_as, self.mygui.menu_item_save_as)
        glal.bind(self.preferences, self.mygui.preferences)
        glal.bind(self.ser1, self.mygui.menu_item_s1)
        glal.bind(self.ser2, self.mygui.menu_item_s2)
        glal.bind(self.ser3, self.mygui.menu_item_s3)
        glal.bind(glal.signalbug, self.mygui.menu_item_bug)
        glal.bind_close(self.exit, self.mygui)

        # Toolbar
        glal.ToolbarBind(self.add_cp, self.mygui.action_add)
        glal.ToolbarBind(self.edit_cp, self.mygui.action_edit)
        glal.ToolbarBind(self.delete_cp, self.mygui.action_delete)
        glal.ToolbarBind(self.plot_graph, self.mygui.action_plot)
        glal.ToolbarBind(self.map_slope, self.mygui.action_map)
        glal.ToolbarBind(self.info, self.mygui.action_properties)

        glal.Message().subscribe(self.slope_add, self.slopelist.message, "SLOPE_ADD")
        glal.Message().subscribe(self.slope_del, self.slopelist.message, "SLOPE_DEL")
        self.mygui.show()

        # Check if this is the latest version
        if self.config.getboolean("main window", "latestVersionCheck"):
            def notify():
                self.versionchk.stop()
                if self.lvq.islast():
                    return
                glal.addstatusbartext(self.mygui, _("A new version of CycloGraph is out!")+"(%s)" % self.lvq.lastversion)
                del self.versionchk
                del self.lvq

            self.lvq = LastVersionQuery()
            self.lvq.start()
            self.versionchk = glal.Timer(10000, notify)  # wait 1s before notify
            self.versionchk.start()

    def init_configparser(self):
        """Create a new configparser, and inizializes with default values,
        if the user saved his preferencese, open configuration file and
        use this values
        """
        self.config = configparser.ConfigParser(DEFAULT_CONF)
        self.config.read(get_config_file_path())

        for sec in SECTIONS:
            if not self.config.has_section(sec):
                self.config.add_section(sec)
        if self.config.get("global", "version") > VERSION:
            # The configuration file has been saved by a future version of
            # CycloGraph thus it is not guaranteeted to be backward compatible.
            # It is better to leave the file as it is and use default settings.
            self.defaultpref()
            return
        if self.config.get("global", "version") < '1.1':
            glal.addstatusbartext(self.mygui,
                                 _("old cofiguration issue."))

    def setguiopt(self):
        """Set gui options to the value from preferences"""
        serv = self.config.get("kml", "server")
        glal.OptionCheck(self.mygui.menu_item_s1, True)
        glal.OptionCheck(self.mygui.menu_item_s2, serv == "earthtools.org")
        glal.OptionCheck(self.mygui.menu_item_s3, serv == "usgs.net")
        altitude_downloader.choose_service(serv)

# Functions responding to menu events ###

    def new_slope(self, event=None):
        """ Create new slope"""
        pagen = self.slopelist.new_slope()
        self.slopelist.set_author(pagen, self.config.get("global", "authorname"))
        self.slopelist.set_email(pagen, self.config.get("global", "authormail"))
        pag = self.notebook.Page()
        pag.slopecounternew = self.slopecounter
        self.notebook.add_page(pag, _('slope')+" "+str(self.slopecounter))
        self.slopecounter = self.slopecounter + 1
        glal.enable_saving(self.mygui, True)

    def open_slope(self, event=None, filepath=None):
        """ Open a file"""
        # statusok = True
        types = [('Supported files', '*.cgx *.csv *.xml *.gpx *.kml *.fitlog *.tcx *.crp *.sal *.txt'),
                    ('CycloGraphXml (*.cgx)', '*.cgx'),
                    ('CycloGraph (*.csv)', '*.csv'),
                    ('Ciclomaniac (*.xml)', '*.xml'),
                    ('GPS eXchange Format (*.gpx)', '*.gpx'),
                    ('Keyhole Markup Language (*.kml)', '*.kml'),
                    ('SportTracks Fitlog files (*.fitlog)', '*.fitlog'),
                    ('Training Center xml (*.tcx)', '*.tcx'),
                    ('Ciclotour (*.crp)', '*.crp'),
                    ('Salitaker (*.sal)', '*.sal'),
                    ('Ciclomaniac (*.txt)', '*.txt')]
        if not filepath:
            filepath = glal.FileDialog(self.mygui, _("Choose a file"), '', "",
                                    types, glal.FD_OPEN)
        if not filepath:
            return
        for nump in range(len(self.slopelist)):
            # Check if the file has already been opened.
            page = self.notebook.get_page(nump)
            if page.savedfilepath == filepath:
                # Select the page of the selected file.
                self.notebook.setselpagnum(nump)
                return
        slope_num = self.slopelist.new_slope()
        pag = self.notebook.Page()
        pag.savedfilepath = filepath
        self.notebook.add_page(pag, "")

        slopefile = None
        filetype = "None"

        if filepath.lower().endswith(".gpx"):
            slopefile = iofile.load_gpxfile(filepath)
            filetype = "gpx"
        elif filepath.lower().endswith(".kml"):
            slopefile = iofile.load_kmlfile(filepath)
            filetype = "kml"
        elif filepath.lower().endswith(".fitlog"):
            slopefile = iofile.load_fitlogfile(filepath)
            filetype = "fitlog"
        elif filepath.lower().endswith(".tcx"):
            slopefile = iofile.load_tcxfile(filepath)
            filetype = "tcx"

        if filetype == "None":
            statusok = iofile.open_file(filepath, self.slopelist, slope_num)
        else:
            if slopefile is not None:
                max_num_cps = len(slopefile)
                (seltype, num, start_dist, end_dist) = glal.gui.numcpsdlg(max_num_cps,
                                                    slopefile.min_dist(),
                                                    slopefile.max_dist(),
                                                    filetype)
                if (seltype == -1):
                    nbook = None
                    self.slopelist.del_slope(slope_num)
                    self.notebook.remove_page(slope_num)
                    return
                if not slopefile.hasAltitudes:
                    self.import_altitudes(num, seltype, slopefile, start_dist, end_dist, slope_num)
                    return
                statusok = slopefile.newSlope(self.slopelist, slope_num, num, start_dist, end_dist, seltype)
                # listcoords = self.slopelist.get_coords(slope_num)
                # print(listcoords)
            else:
                statusok = False

        if not statusok:
            nbook = None
            self.slopelist.del_slope(slope_num)
            self.notebook.remove_page(slope_num)
            glal.addstatusbartext(self.mygui, _("Unable to open file"))
            return
        self.tabupdate(slope_num)
        glal.enable_saving(self.mygui,  True)
        if statusok == 2:
            glal.addstatusbartext(self.mygui,
                                  _("Old file format: please save it again."))

    def create_kml(self, event=None):
        """ Create kml file"""
        # TODO: check for internet connection
        def downloadfunc(url):
            """Download current create_kml page"""
            from urllib.request import urlopen
            if url:
                response = urlopen(url)
                return response

        def savefunc(content, mode):
            """Save as kml current create_kml selected route"""
            if mode != 'DRAW' and mode != 'YOURS':
                content = iofile.gpxtokml(content)
            filepath = glal.FileDialog(None, _("kml name"), "", "",
                   [("Keyhole Markup Language (*.kml)", "*.kml")], glal.FD_SAVE)
            if not filepath:
                return
            if not filepath.endswith('.kml'):
                filepath = filepath+".kml"
            with open(filepath, "w") as outfile:
                print(content, file=outfile)

        create = glal.gui.Create_kml(self.mygui, downloadfunc, savefunc)
        accepted = create.show_modal()
        if not accepted:
            create.destroy()
            return
        if not create.content:
            glal.addstatusbartext(self.mygui,
            _("Network error : unable to create an handler to download kml route."))
            return
        if create.mode == 'DRAW':
            # If the user drawned the path cg must use every point,
            # and not waste user effort.
            numcps = 99999  # Hack: unlikely any user will draw more points.
            kmlcontent = create.content
        else:
            numcps = self.config.getint("kml", "numcps")
        if create.mode == 'ORS':
            gpxcontent = create.content
            kmlcontent = iofile.gpxtokml(gpxcontent)
        elif create.mode == 'BROU':
            gpxcontent = create.content
            kmlcontent = iofile.gpxtokml(gpxcontent)
        elif create.mode == 'YOURS':
            kmlcontent = create.content
        kmlfile = iofile.load_kmlfile(kmlcontent)
        if kmlfile is None:
            glal.addstatusbartext(self.mygui, _("Error on loading slope"))
            create.destroy()
            return
        slope_num = self.slopelist.new_slope()
        nbook = self.notebook.Page()
        nbook.slopecounternew = self.slopecounter
        self.notebook.add_page(nbook, _('slope %d') % self.slopecounter)
        self.slopecounter = self.slopecounter + 1
        self.tabupdate(slope_num, '*')

        self.import_altitudes(numcps, 0, kmlfile,
                                         kmlfile.min_dist(),
                                         kmlfile.max_dist(),
                                         slope_num)
        create.destroy()

    def import_altitudes(self, num, seltype, slopefile, start_dist, end_dist, slope_num):
        """ Start the import kml process in CycloGraph"""
        # TODO: check for internet connection
        self.slopenumber_kml = slope_num
        slopefile.getCoords(self.slopelist, slope_num, start_dist, end_dist)
        self.pdlg = glal.ProgressDialog()
        refresh_time = 30  # FIXME: is the measure unit milliseconds?
        self.timer = glal.Timer(refresh_time, self.on_altitude_importing)
        self.queue = queue.Queue()
        self.prsrthrd = altitude_downloader.ImporterThread(self.queue, slopefile, num, start_dist, end_dist, seltype)
        self.prsrthrd.daemon = True
        # start thread and timer
        self.prsrthrd.start()
        self.timer.start()

    def on_altitude_importing(self, widget=None):
        """ Called by a timer:
        check if thread is still executing
        (if not stop timer and, according to status, handle the produced slope)
        reads the progress of importer-thread
        updates the progress dialog
        check if operation is cancelled and tells to thread
        """
        try:
            (progress, chkp) = self.queue.get(block=False)
        except queue.Empty:
            # print("is alive "+str(self.prsrthrd.isAlive()))
            if not self.prsrthrd.isAlive():
                self.timer.stop()
                self.pdlg.destroy()
                if self.prsrthrd.status != 'OK':
                    slope_num = self.slopenumber_kml
                    self.slopelist.del_slope(slope_num)
                    self.notebook.remove_page(slope_num)
                    glal.addstatusbartext(self.mygui,
                                          _("Error while loading data"))
                else:
                    self.tabupdate(self.slopenumber_kml, '*')
                    glal.enable_saving(self.mygui, True)
                del self.prsrthrd
            # there is nothing to update but gtk wants True
            return True
        self.slopelist.add_cp(self.slopenumber_kml, *chkp)
        # There is a new progress value => update the gui
        want2abort = self.pdlg.update(progress)
        if want2abort:
            self.prsrthrd.abort()
        # need to return True to keep timer_gtk running
        return True

    def save_slope_as(self, event=None, pagenumber=None, filename=None):
        """ Save_as selected slope """
        if pagenumber is None:
            pagenumber = self.notebook.getselpagnum()
        myslope = self.slopelist.get_slope_copy(pagenumber)
        filepath = glal.FileDialog(self.mygui, _("Save as"), '', "",
            [("CycloGraphXml(*.cgx)", "*.cgx")], glal.FD_SAVE, filename)
        if not filepath:
            return False
        if not filepath.endswith('.cgx'):
            filepath = filepath + '.cgx'
        iofile.save_file(filepath, myslope)
        page = self.notebook.get_page(pagenumber)
        page.savedfilepath = filepath
        self.tabupdate(pagenumber)
        return True

    def save_slope(self, event=None):
        """ Save selected slope """
        pagenumber = self.notebook.getselpagnum()
        page = self.notebook.get_page(pagenumber)
        myslope = self.slopelist.get_slope_copy(pagenumber)
        filepath = page.savedfilepath
        if filepath is None or not filepath.endswith('.cgx'):
            filepath = glal.FileDialog(self.mygui, _("Choose a file"), '', "",
                [("CycloGraphXml(*.cgx)", "*.cgx")], glal.FD_SAVE)
            if not filepath:
                return
        if not filepath.endswith('.cgx'):
            filepath = filepath + '.cgx'
        iofile.save_file(filepath, myslope)
        page.savedfilepath = filepath
        self.tabupdate(pagenumber)
        glal.addstatusbartext(self.mygui, _("File saved"))

    def exit(self):
        """ Check if there are unsaved slopes, and cg is ready to exit
        This funciotn is called by the gui"""
        if not len(self.slopelist):
            if not self.mygui.ismaximized():
                dim = self.mygui.getdimensions()
                self.config.set("main window", "widthsize", str(dim[0]))
                self.config.set("main window", "heightsize", str(dim[1]))
            self.config.set("main window", "maximized",
                            "%s" % self.mygui.ismaximized())
            self.savepref()
            return True
        for pagenum in reversed(range(len(self.slopelist))):
            self.check_unsaved(pagenum)
        if not len(self.slopelist):
            if not self.mygui.ismaximized():
                dim = self.mygui.getdimensions()
                self.config.set("main window", "widthsize", str(dim[0]))
                self.config.set("main window", "heightsize", str(dim[1]))
            self.config.set("main window", "maximized",
                            "%s" % self.mygui.ismaximized())
            self.savepref()
            return True
        else:
            return False

    def preferences(self, event=None):
        config = configparser.ConfigParser(DEFAULT_CONF)
        config.read(get_config_file_path())

        for sec in SECTIONS:
            if not config.has_section(sec):
                config.add_section(sec)
        if config.get("global", "version") > VERSION:
            self.defaultpref()
        settings = getconfig(config)
        pref = glal.gui.preferences(self.mygui, settings, DEFAULT_CONF)
        if not pref:
            return
        self.config.set("global", "orizzontal lines", str(pref['olines']))
        self.config.set("global", "legend", str(pref['legend']))
        self.config.set("global", "theme", pref['theme'])
        self.config.set("global", "slopeinfo", str(pref['sinfo']))
        self.config.set("global", "effect-3d", str(pref['3d']))
        self.config.set("global", "latestVersionCheck", str(pref['vcheck']))
        self.config.set("global", "authorname", str(pref['aname']))
        self.config.set("global", "authormail", str(pref['amail']))
        font = pref['fdesc']
        self.config.set("font", "font-des 1", font["des"])
        self.config.set("font", "font-typ 1", font["typ"])
        self.config.set("font", "font-dim 1", str(font["dim"]))
        font = pref['ftitle']
        self.config.set("font", "font-des t", font["des"])
        self.config.set("font", "font-typ t", font["typ"])
        self.config.set("font", "font-dim t", str(font["dim"]))
        font = pref['fgrad']
        self.config.set("font", "font-des g", font["des"])
        self.config.set("font", "font-typ g", font["typ"])
        self.config.set("font", "font-dim g", str(font["dim"]))
        self.config.set("resolution", "width", pref['width'])
        self.config.set("resolution", "height", pref['height'])
        self.config.set("kml", "server", pref['serv'])
        colorlist = pref['colors']
        levellist = pref['levels']
        for i in range(0, 7):
            self.config.set("colors", "color "+str(i+1), str(colorlist[i]))
            self.config.set("levels", "level "+str(i+1), str(levellist[i]))
        self.savepref()
        self.setguiopt()


# Functions related to kml server ###
    def ser1(self, event=None):
        """Set kml server1"""
        self.config.set("kml", "server", "geonames.org")
        altitude_downloader.choose_service("geonames.org")

    def ser2(self, event=None):
        """Set kml server2"""
        self.config.set("kml", "server", "earthtools.org")
        altitude_downloader.choose_service("earthtools.org")

    def ser3(self, event=None):
        """Set kml server3"""
        self.config.set("kml", "server", "usgs.net")
        altitude_downloader.choose_service("usgs.net")

    def savepref(self, event=None):
        """Save preferences.
        Configuration file is save in the user's home
        as standardized byXDG."""
        with open(get_config_file_path(), "w") as fid:
            self.config.write(fid)

    def defaultpref(self, event=None):
        """Load default preferences: delete current config parser
        and recreate a new configparser with default values"""
        del self.config
        self.config = configparser.ConfigParser(DEFAULT_CONF)
        for sec in SECTIONS:
            if not self.config.has_section(sec):
                self.config.add_section(sec)
        self.setguiopt()

# Functions related to a single Page ###

    def close(self, *argument):  # only qt < 4.5 doesn't pass any argument
        """Close a single page of notebook."""
        if not len(self.slopelist):
            return
        if not argument:
            pagenum = self.notebook.getselpagnum()
        else:
            pagenum = self.notebook.get_pagenum(argument)
        self.check_unsaved(pagenum)

    def tabupdate(self, num, tag=''):
        """Update page.modified and title of the page tab"""
        page = self.notebook.get_page(num)
        name = self.slopelist.get_name(num)
        if tag == '':
            page.modified = False
        else:
            page.modified = True

        if name == '' or name is None:
            if page.savedfilepath is not None:
                self.notebook.set_page_label(num, tag+os.path.split(page.savedfilepath)[-1])
            else:
                self.notebook.set_page_label(num,  tag+_('slope')+" "+str(page.slopecounternew))
            return
        self.notebook.set_page_label(num,  tag+name)

    def check_unsaved(self, pagenum):
        """ Check if passed page is not saved and ask to save it"""
        page = self.notebook.get_page(pagenum)
        if page.modified:
            if page.savedfilepath is None:
                response = glal.gui.save_changes(self.mygui, _('slope') + " " +
                                                str(page.slopecounternew))
                if (response == 'SAVE'
                        and not self.save_slope_as(None, pagenum)) \
                        or response != 'DISCARD':
                    return
            else:
                filename = os.path.split(page.savedfilepath)[-1]
                response = glal.gui.save_changes(self.mygui, filename)
                if response == 'SAVE':
                    if not filename.endswith('.cgx'):
                        self.save_slope_as(None, pagenum,
                                                 filename.rsplit('.', 1)[0]+'.cgx')
                        return
                    filepath = page.savedfilepath
                    myslope = self.slopelist.get_slope_copy(pagenum)
                    iofile.save_file(filepath, myslope)
                    page.savedfilepath = filepath
                    page.modified = False
                elif not response == 'DISCARD':
                    return
        self.notebook.remove_page(pagenum)
        self.slopelist.del_slope(pagenum)
        if not len(self.slopelist):
            glal.enable_saving(self.mygui, False)

    def plot_graph(self, event=None):
        """Open the Plot windows showing a graphical image of current slope"""
        # slope_num == pagenumber == num
        num = self.notebook.getselpagnum()
        myslope = self.slopelist.get_slope_copy(num)
        if len(myslope) < 2:
            glal.gui.geterrnocp(self.mygui)
            return
        plotdata = getconfig(self.config)
        myplot = glal.gui.Plot(self.mygui, myslope,
                                plotdata, export_as_image, save_pdf, save_html)
        myplot.showplot()

    def map_slope(self, event=None):
        """Open a Map window"""
        num = self.notebook.getselpagnum()
        slope = self.slopelist.get_slope_copy(num)
        if len(slope.coords) != 0:
            (lat, lng) = slope.coords[0]
            stringc = "%s %s" % (lng, lat)
            for i in range(1, len(slope.coords)):
                (lat, lng) = slope.coords[i]
                stringc += ",%s %s" % (lng, lat)
            glal.gui.Map(self.mygui, slope.name, stringc, save_map_image)
        else:
            glal.gui.mapcoorderr(self.mygui)

    def add_cp(self, event=None):
        """Add a check point to current slope."""
        num = self.notebook.getselpagnum()
        mng = glal.gui.Managecp(self.mygui, _("Add"), "", "", "")
        if mng.show_modal():
            (dist, alt, name) = mng.getcp()
            if (dist != "") and (alt != ""):
                self.slopelist.add_cp(num, float(dist), float(alt), name)
                self.tabupdate(num, '*')
            else:
                glal.gui.managecperr(self.mygui)
        mng.destroy()

    def edit_cp(self, event=None):
        """Edit a check point to current slope."""
        num = self.notebook.getselpagnum()
        page = self.notebook.get_page(num)
        editcp = page.get_sel_row()
        myslope = self.slopelist.get_slope_copy(num)
        if editcp < 0:
            return
        (dist, alt, name) = (str(myslope.cps[editcp][0]),
                             str(myslope.cps[editcp][1]),
                             str(myslope.cps[editcp][2]))
        mng = glal.gui.Managecp(self.mygui, _("Edit"), dist, alt, name)
        if mng.show_modal():
            (dist, alt, name) = mng.getcp()
            if (dist != "") and (alt != ""):
                self.slopelist.remove_cp(num, editcp)
                self.slopelist.add_cp(num, float(dist), float(alt), name)
                page = self.notebook.get_page(num)
                self.tabupdate(num, '*')
            else:
                glal.gui.managecperr(self.mygui)
        mng.destroy()

    def delete_cp(self, event=None):
        """Delete a check point to current slope."""
        num = self.notebook.getselpagnum()
        page = self.notebook.get_page(num)
        rm_cps = page.get_sel_multi_row()
        # remove selected cps starting from the end
        # to not affect row number of non removed items
        rm_cps.reverse()
        for i in rm_cps:
            self.slopelist.remove_cp(num, i)
        self.tabupdate(num, '*')

    def info(self, event=None):
        """Show informations on current slope"""
        num = self.notebook.getselpagnum()
        # need to call calc to have the correct value of average and max gradient
        myslope = self.slopelist.get_slope_copy(num)
        myslope.calc()
        slopedata = {'name': self.slopelist.get_name(num),
                     'state': self.slopelist.get_state(num),
                     'author': self.slopelist.get_author(num),
                     'email': self.slopelist.get_email(num),
                     'comment': self.slopelist.get_comment(num),
                     'average_grad': "%.1f" % self.slopelist.get_average_grad(num)+' %',
                     'max_grad': "%.1f" % self.slopelist.get_max_grad(num)+' %',
                     'height_difference': str(self.slopelist.get_height_difference(num)) + ' m',
                     'height_gain': str(self.slopelist.get_height_gain(num)) + ' m',
                     }
        dlg = glal.gui.formslopeinfo(self.mygui, slopedata)
        if not dlg:
            return
        self.slopelist.set_name(num, dlg['name'])
        self.slopelist.set_country(num, dlg['state'])
        self.slopelist.set_author(num, dlg['author'])
        self.slopelist.set_email(num, dlg['email'])
        self.slopelist.set_comment(num, dlg['comment'])
        self.tabupdate(num, '*')

# Functions activated in response to Messages by the GUI ###

    def slope_add(self, pagenumber, row_num):
        """Callback from a SLOPE ADD message"""
        page = self.notebook.get_page(pagenumber)
        if page is None:
            return
        myslope = self.slopelist.get_slope_copy(pagenumber)
        page.insert_row(row_num, myslope.cps[row_num])

    def slope_del(self, pagenumber, row_num):
        """Callback from a SLOPE DEL message"""
        page = self.notebook.get_page(pagenumber)
        if page is None:
            return
        page.delete_row(row_num)


# Functions related to Map ###
def save_map_image(mapwindow):
    """ Save map to an image"""
    formats = [("Portable Network Graphics (*.png)", "*.png"),
               ("Bitmap (*.bmp)", "*.bmp"),
               ("Joint Photographic Experts Group (*.jpg)", "*.jpg")]
    filepath = ''
    # Control if the right extension is given by user
    while not filepath.lower().endswith(('.png', '.bmp', '.jpg')):
        if filepath:
            filepath += '.png'
        filepath = glal.FileDialog(None, _("Save map"), filepath,
                                   "", formats, glal.FD_SAVE)
        if not filepath:
            return
    mapwindow.saveimage(filepath)


# Functions related to Plot ###
def export_as_image(imgslope, settings):
    """ Save a plot to an image"""
    formats = [("Portable Network Graphics (*.png)", "*.png"),
               ("Bitmap (*.bmp)", "*.bmp"),
               ("Joint Photographic Experts Group (*.jpg)", "*.jpg"),
               ("Scalable Vector Graphics (*.svg)", "*.svg")]
    filepath = ''
    # Control if the right extension is given by user
    while not filepath.lower().endswith(('.png', '.bmp', '.jpg', '.svg')):
        if filepath:
            filepath += '.png'
        filepath = glal.FileDialog(None, _("Save plot"), filepath,
                                   "", formats, glal.FD_SAVE)
        if not filepath:
            return

    format = filepath.rsplit('.', 1)[1]
    if format.lower() == 'svg':
        image = glal.Image_svg(settings['width'], settings['height'],
                               imgslope.paint)
    else:
        image = glal.Image(settings['width'], settings['height'],
                           imgslope.paint)
    image.plot(settings)
    image.savetofile(filepath, format)


def save_pdf(imgslope, settings):
    """ Save a pdf from current plot"""
    formats = [("Portable Document Format (*.pdf)", "*.pdf")]
    filepath = ''
    # Check if the right extension is given by user
    while not filepath.lower().endswith('.pdf'):
        if filepath:
            filepath += '.pdf'
        filepath = glal.FileDialog(None, _("Save plot to pdf"), filepath,
                                   "", formats, glal.FD_SAVE)
        if not filepath:
            return

    pdf = glal.Pdf(filepath)
    pdfconfig = settings.copy()
    pdfconfig['sinfo'] = False
    pdf.plot_image(pdfconfig, 793, 600,
                   # settings['width'], settings['height'],)
                   imgslope.paint)
    pdf.addtitle(_("Slope"))
    pdf.addtext(_("Name") + ": " + imgslope.name)
    pdf.addtext(_("Country") + ": " + imgslope.country)
    pdf.addtext(_("Average gradient") + ": " + "%.1f" % imgslope.average_grad+" %")
    pdf.addtext(_("Max gradient") + ": " + "%.1f" % imgslope.max_grad+" %")
    pdf.addtext(_("Height difference") + ": " + str(imgslope.height_difference)+" m")
    pdf.addtext(_("Height gain") + ": " + str(imgslope.height_gain)+" m")
    pdf.addtitle(_("Author"))
    pdf.addtext(_("Name") + ": " + imgslope.author)
    pdf.addtext(_("E-mail") + ": " + imgslope.email)
    # TODO:use correct indentation to include notes on the pdf output
    # pdf.addtitle(_("Note"))
    # pdf.addtext(imgslope.comment)
    pdf.save()


def save_html(imgslope, settings):
    """ Save an html from current plot"""
    formats = [("HyperText Markup Language (*.html)", "*.html")]
    filepath = ''
    # Check if the right extension is given by user
    while not filepath.lower().endswith('.html'):
        if filepath:
            filepath += '.html'
        filepath = glal.FileDialog(None, _("Save plot to html"), filepath,
                                   "", formats, glal.FD_SAVE)
        if not filepath:
            return

    image = glal.Image_svg(settings['width'], settings['height'], imgslope.paint)
    image.plot(settings)

    reportpage = report_html.Report_html(imgslope.name)
    reportpage.add_image(image.svg)
    if len(imgslope.coords) != 0:
        reportpage.add_map(imgslope, settings['width'], settings['height'])
    reportpage.addtitle(_("Slope"))
    reportpage.addtext(_("Name")+": ", imgslope.name)
    reportpage.addtext(_("Country")+": ", imgslope.country)
    reportpage.addtext(_("Average gradient")+": ", "%.1f" % imgslope.average_grad+" %")
    reportpage.addtext(_("Max gradient")+": ", "%.1f" % imgslope.max_grad+" %")
    reportpage.addtext(_("Height difference")+": ", str(imgslope.height_difference)+" m")
    reportpage.addtext(_("Height gain")+": ", str(imgslope.height_gain)+" m")
    reportpage.addtitle(_("Author"))
    reportpage.addtext(_("Name")+": ", imgslope.author)
    reportpage.addtext(_("E-mail")+": ", imgslope.email)
    if imgslope.comment != '':
        reportpage.addtitle(_("Note")+": ")
        reportpage.addnote(imgslope.comment)
    reportpage.save(filepath)


def get_config_file_path():
    config_path = os.path.expanduser(os.path.normpath('~/.config/CycloGraph'))
    if 'XDG_CONFIG_HOME' in os.environ and os.path.isabs(os.environ['XDG_CONFIG_HOME']):
        config_path = os.environ['XDG_CONFIG_HOME'] + os.path.normpath('/CycloGraph')
    config_file = config_path + os.path.normpath('/CycloGraph.conf')

    # Create a directory if needed
    if not os.path.exists(config_path):
        os.makedirs(config_path)
        # migrate old configuration file if there is any
        old_file = os.path.expanduser(os.path.normpath('~/.cyclograph.cfg'))
        if os.path.exists(old_file):
            os.rename(old_file, config_file)

    return config_file


def getconfig(config):
    """ Return current settings from config"""
    settings = {
                'olines': config.getboolean("global", "orizzontal lines"),
                'legend': config.getboolean("global", "legend"),
                'theme': config.get("global", "theme"),
                'sinfo': config.getboolean("global", "slopeinfo"),
                '3d': config.getboolean("global", "effect-3d"),
                'vcheck': config.getboolean("global", "latestVersionCheck"),
                'fdesc': {
                          "des": config.get("font", "font-des 1"),
                          "typ": config.get("font", "font-typ 1"),
                          "dim": config.getint("font", "font-dim 1")
                          },
                'ftitle': {
                           "des": config.get("font", "font-des t"),
                           "typ": config.get("font", "font-typ t"),
                           "dim": config.getint("font", "font-dim t")
                           },
                'fgrad': {
                          "des": config.get("font", "font-des g"),
                          "typ": config.get("font", "font-typ g"),
                          "dim": config.getint("font", "font-dim g")
                          },
                'width': config.getint("resolution", "width"),
                'height': config.getint("resolution", "height"),
                'themelist': ThemeManager().getthemeslist(),
                'serv': config.get("kml", "server"),
                'aname': str(config.get("global", "authorname")),
                'amail': str(config.get("global", "authormail")),
                'colors': [str(config.get("colors",
                    "color "+str(i))) for i in range(1, 8)],
                'levels': [config.getfloat("levels",
                    "level "+str(i)) for i in range(1, 8)]
                }
    return settings

########################################################


def usage():
    """ Print usage on standard error"""
    sys.stderr.write("usage: %s {-q|-G|-f}\n" % sys.argv[0])
    sys.stderr.write("\t--qt --gtk3 or --file=*.cgx file for cli\n")


def opt_parser(argv):
    """ Parse console input"""
    domain = gettext.textdomain(None)
    gettext.textdomain("help")
    parser = argparse.ArgumentParser(prog="CycloGraph", description="CycloGraph "+VERSION)

    parser.set_defaults(filename="")
    parser.set_defaults(gui="cli")
    parser.add_argument("-v", "--version", action="version", version="%(prog)s "+VERSION)
    parser.add_argument("-G", "--gtk3", help="gtk3 graphical user interface", action="store_const", dest="gui", const="gtk3")
    parser.add_argument("-q", "--qt", help="qt graphical user interface", action="store_const", dest="gui", const="qt")
    parser.add_argument("-f", "--file", dest="filename", help="open cgx, csv, gpx, tcx, sal, crp or cyclomaniac file", metavar="FILE")
    parser.add_argument("-s", "--size", dest="size", action="store", help="output size widthxheight")
    parser.add_argument("-l", "--list", dest="_list", action="store_true", help="list available themes and exit")
    parser.add_argument("-o", "--options", action='append', dest='optionslist', default=[], help="add option to optionlist", choices=['3d', 'legend', 'olines', 'sinfo'])
    parser.add_argument("-t", "--theme", action="store", dest="theme", help="use selected theme")
    parser.add_argument("-e", "--extension", dest="extension", default="", help="choose default file format extension", metavar="EXT")
    options = parser.parse_args()
    if options._list:
        # print("available themes are:")
        # User asked for the available themes, so telling him is quite useless
        # also is less machine readable: silence is golden.
        for theme in ThemeManager().getthemeslist():
            print(theme)
        sys.exit(0)
    if options.theme and \
       options.theme not in ThemeManager().getthemeslist():
        parser.error('there is no theme named "%s"' % options.theme)
    gettext.textdomain(domain)
    return options

########################################################


def main_qt():
    """Main application for qt interface"""
    app = glal.initapp(sys.argv)

    arg = [str(s) for s in app.arguments()]
    options = opt_parser(arg)

    # WARNING: The return value of CycloGraph MUST be stored in a variable
    # otherwise the application could segfault.
    controller = Cyclograph()
    if options.filename:
        controller.open_slope(filepath=options.filename)
    sys.exit(app.exec_())


def main_gtk3():
    """Main application for gtk3 interface"""
    controller = Cyclograph()
    options = opt_parser(sys.argv)
    if options.filename:
        controller.open_slope(filepath=options.filename)
    glal.initapp()


def main_cli():
    """Main application for cli interface"""
    slopelist = slope.SlopeList()
    slope_num = slopelist.new_slope()
    opt = opt_parser(sys.argv)
    filein = opt.filename
    if filein == "-":
        filein = sys.stdin
    success = iofile.open_file(filein, slopelist, slope_num, opt.extension)
    if not success:
        print(('Unable to open "%s"' % filein), file=sys.stderr)
        return False
    devc = glal.DeviceContext()
    myslope = slopelist.get_slope_copy(slope_num)
    myslope.calc()

    config = configparser.ConfigParser(DEFAULT_CONF)
    config.read(get_config_file_path())

    for sec in SECTIONS:
        if not config.has_section(sec):
            config.add_section(sec)
    settings = getconfig(config)
    if opt.theme:
        settings['theme'] = opt.theme  # FIXME: this should work also with GUIs
    size_w = settings['width']
    size_h = settings['height']
    if opt.optionslist:
        if "3d" in opt.optionslist:
            settings['3d'] = True
        else:
            settings['3d'] = False
        if "legend" in opt.optionslist:
            settings['legend'] = True
        else:
            settings['legend'] = False
        if "olines" in opt.optionslist:
            settings['olines'] = True
        else:
            settings['olines'] = False
        if "sinfo" in opt.optionslist:
            settings['sinfo'] = True
        else:
            settings['sinfo'] = False
    if opt.size:
        size_el = opt.size.split('x')
        if len(size_el) == 2:
            try:
                el_w = int(size_el[0])
                el_h = int(size_el[1])
                if (el_w > 0) and (el_h > 0):
                    size_w = el_w
                    size_h = el_h
            except ValueError:
                print(('ValueError reading size "%s"' % opt.size), file=sys.stderr)
    devc.init_draw_surf((size_w, size_h), settings['3d'])
    myslope.paint(settings, devc)
    devc.end_draw()

# vim:sw=4:softtabstop=4:expandtab



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""launcher.py """

# Copyright (C) 2008-2014 Federico Brega, Pierluigi Villani

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import sys
import locale
import gettext
import os
import glal_selected


def setup_gettext():
    """Application translation system setup"""

    # Setup translation.
    langs = []
    mylocale = locale.getdefaultlocale()[0]
    if mylocale:
        langs = [mylocale]
    # Other languages in the system.
    language = os.environ.get('LANGUAGE', None)
    if language:
        langs += language.split(':')
    # If there isn't any better use this.
    langs += ["en_CA", "en_US"]
    po_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                           '..', 'po')

    if not os.path.isdir(po_path):
        po_path = "/usr/share/locale"
    gettext.bindtextdomain("cyclograph", po_path)
    gettext.textdomain("cyclograph")
    lang = gettext.translation("cyclograph", po_path,
                                     languages=langs, fallback=True)
    lang.install()


def main():
    setup_gettext()
    if len(sys.argv) < 2:
        # If there has been an error with the commandline specified frontend
        # try others in the order (PyQt gets first because is the supported
        # frontend on Mac Os and Windows where users might want a graphical
        # interface.
        if glal_selected.use('qt'):
            import cyclograph
            cyclograph.main_qt()
        elif glal_selected.use('gtk3'):
            import cyclograph
            cyclograph.main_gtk3()
        elif glal_selected.use('cli'):
            import cyclograph
            cyclograph.main_cli()
        else:
            sys.exit(2)
    else:
        # also file=* is an argument even if it is not space separed.
        arguments = set([token.partition('=')[0] for token in sys.argv])
        if arguments & set(('-q', '--qt')):
            if glal_selected.use('qt') != 'qt':
                sys.stderr.write('Qt backend is not installed correctely\n')
                sys.exit(3)
            import cyclograph
            cyclograph.main_qt()
        elif arguments & set(('-G', '--gtk3')):
            if glal_selected.use('gtk3') != 'gtk3':
                sys.stderr.write('GTK+ 3 backend is not installed correctely\n')
                sys.exit(3)
            import cyclograph
            cyclograph.main_gtk3()
        elif arguments & set(('-f', '--file', '-e', '--extension')):
            if glal_selected.use('cli') != 'cli':
                sys.stderr.write('Fatal error, cli backend cannot be started\n')
                sys.exit(3)
            import cyclograph
            cyclograph.main_cli()
        else:
            glal_selected.use('cli')
            import cyclograph
            cyclograph.main_cli()
            sys.exit(1)

if __name__ == '__main__':
    main()

# vim:sw=4:softtabstop=4:expandtab



# -*- coding: utf-8 -*-
"""iofile.py"""

# Copyright (C) 2008-2017 Federico Brega, Pierluigi Villani

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

from __future__ import with_statement
import csv
import logging
import os
import xml.dom.minidom
import zlib
from altitude_downloader import distance, haversine, point_conversion

CSV_VERSION = "1.1"
CGX_VERSION = "1.0"

STAT_VER = 2


def save_file(filename, slope):
    """Save .cgx file"""
    doc = xml.dom.minidom.Document()
    cg = doc.createElement(u"cyclographxml")
    cg.setAttribute(u"version", CGX_VERSION)
    doc.appendChild(cg)
    sl = doc.createElement(u"slope")
    cg.appendChild(sl)

    sl.appendChild(__createTextElement(doc, u"name", slope.name))
    sl.appendChild(__createTextElement(doc, u"country", slope.country))

    author = doc.createElement(u"author")
    author.appendChild(__createTextElement(doc, u"name", slope.author))
    author.appendChild(__createTextElement(doc, u"email", slope.email))
    sl.appendChild(author)

    sl.appendChild(__createTextElement(doc, u"info", slope.comment))

    cps = doc.createElement(u"cps")
    for cpt in slope.cps:
        cp = doc.createElement(u"cp")
        (dist, alt, desc) = cpt
        cp.setAttribute(u"alt", "%.3f" % alt)
        cp.setAttribute(u"dist", "%.3f" % dist)
        if desc != "":
            cp.setAttribute(u"desc", desc)
        cps.appendChild(cp)
    sl.appendChild(cps)

    coords = doc.createElement(u"coords")
    for coor in slope.coords:
        coord = doc.createElement(u"coord")
        (lat, lon) = coor
        coord.setAttribute(u"lon", "%.6f" % lon)
        coord.setAttribute(u"lat", "%.6f" % lat)
        coords.appendChild(coord)
    sl.appendChild(coords)

    with open(filename, "w") as fid:
        doc.writexml(fid, encoding='utf-8', newl='\n')


def __createTextElement(doc, name, text):
    elem = doc.createElement(name)
    ptext = doc.createTextNode(text)
    elem.appendChild(ptext)
    return elem

# def save_file(filename, slope):
#    """Save .csv file"""
#    #Control if the right extension is given by user
#    if not filename.endswith('.csv'):
#        filename = filename + '.csv'
#    #fid = open(filename,'wb')
#    # Too bad there is no csv unicode writer so open in ASCII mode and convert utf-8 to ASCII
#    # Be carreful to also encode delimiter and quotechar if they are not ASCII
#    with open(filename, mode='wb') as fid:
#        #fid.write ("%s;%s;%s\n" % (slope.name, slope.country, slope.comment) )
#        csv_writer = csv.writer(fid, delimiter=';', quotechar='\"')
#        def to_unicode(item):
#            if isinstance(item, unicode):
#                item = item.encode('utf-8')
#            return item
#        csv_writer.writerow( map(to_unicode, (str(CSV_VERSION), slope.name, slope.country,
#                               slope.author, slope.email, slope.comment)) )
#        for row in slope.cps:
#            csv_writer.writerow(map(to_unicode, row))
#        #for cpt in slope.cps:
#        #    fid.write ("%.3f;%.3f;%s\n" % cpt)
#    #fid.close()


def open_file(filename, slopelist, slopenum, extension=''):
    """Open generic file"""
    statusok = True
    if not extension:
        if isinstance(filename, str):
            extension = filename.split('.')[-1].lower()
        else:
            extension = 'cgx'  # when used stdin
    if extension == 'cgx':
        try:
            statusok = __open_cgx(filename, slopelist, slopenum)
        except IOError:
            statusok = False
    elif extension == 'csv':
        try:
            statusok = __open_csv(filename, slopelist, slopenum)
        except IOError:
            statusok = False
    elif extension == 'xml':
        try:
            statusok = __open_xml(filename, slopelist, slopenum)
        except IOError:
            statusok = False
    elif extension == 'gpx':
        try:
            gpxfile = load_gpxfile(filename)
            gpxfile.newSlope(slopelist, slopenum, -1, gpxfile.min_dist(), gpxfile.max_dist())
            statusok = True
        except IOError:
            statusok = False
    elif extension == 'kml':
        try:
            kmlfile = load_kmlfile(filename)
            kmlfile.newSlope(slopelist, slopenum, -1, kmlfile.min_dist(), kmlfile.max_dist())
            statusok = True
        except IOError:
            statusok = False
    elif extension == 'fitlog':
        try:
            fitlogfile = load_fitlogfile(filename)
            fitlogfile.newSlope(slopelist, slopenum, -1, fitlogfile.min_dist(), fitlogfile.max_dist())
            statusok = True
        except IOError:
            statusok = False
    elif extension == 'tcx':
        try:
            tcxfile = load_tcxfile(filename)
            tcxfile.newSlope(slopelist, slopenum, -1, tcxfile.min_dist(), tcxfile.max_dist())
            statusok = True
        except IOError:
            statusok = False
    elif extension == 'sal':
        try:
            __open_sal(filename, slopelist, slopenum)
        except IOError:
            statusok = False
    elif extension == 'crp':
        try:
            __open_crp(filename, slopelist, slopenum)
        except IOError:
            statusok = False
    elif extension == 'txt':
        try:
            __open_txt(filename, slopelist, slopenum)
        except IOError:
            statusok = False
    else:
        statusok = False
    return statusok

# Functions defining how to open a certain file format.


def __open_csv(filename, slopelist, slopenum):
    """Open a Cyclograph CSV file (legacy)"""
    if isinstance(filename, str):
        fid = open(filename, encoding='utf-8', mode='r')
    else:
        fid = filename
    csv_reader = csv.reader(fid, delimiter=';', quotechar='\"')
    line = next(csv_reader)
    # Make it more robust: if file has less than the required elements
    # it considers them as empty strings and doesn't consider exceding ones.
    if len(line) < 6:
        line.extend([""] * (6-len(line)))
    (version, slopename, slopestate, slopeauthor, authoremail, slopecomment) = line[:6]
    status = True

    # This is a legacy format so always warn about that, regardless the version number
    status = STAT_VER

    slopelist.set_name(slopenum, slopename)
    slopelist.set_country(slopenum, slopestate)
    slopelist.set_author(slopenum, slopeauthor)
    slopelist.set_email(slopenum, authoremail)
    slopelist.set_comment(slopenum, slopecomment)

    for line in csv_reader:
        # (dist, alt, name)=line.strip().split(';')
        try:
            (dist, alt, name) = line
        except ValueError:
            return False
        slopelist.add_cp(slopenum, float(dist), float(alt), name)
    fid.close()
    return status


def __open_cgx(filename, slopelist, slopenum):
    """Open a Cyclograph cgx file"""
    fid = xml.dom.minidom.parse(filename)
    slop = fid.getElementsByTagName("slope")[0]
    slopename = slop.getElementsByTagName("name")[0]
    if slopename.hasChildNodes():
        slopelist.set_name(slopenum, slopename.childNodes[0].data)
    slopestate = slop.getElementsByTagName("country")[0]
    if slopestate.hasChildNodes():
        slopelist.set_country(slopenum, slopestate.childNodes[0].data)

    auth = slop.getElementsByTagName("author")[0]
    author = auth.getElementsByTagName("name")[0]
    if author.hasChildNodes():
        slopelist.set_author(slopenum, author.childNodes[0].data)
    mail = auth.getElementsByTagName("email")[0]
    if mail.hasChildNodes():
        slopelist.set_email(slopenum, mail.childNodes[0].data)

    slopecomment = slop.getElementsByTagName("info")[0]
    if slopecomment.hasChildNodes():
        slopelist.set_comment(slopenum, slopecomment.childNodes[0].data)

    for cpt in slop.getElementsByTagName('cp'):
        try:
            slopelist.add_cp(slopenum, float(cpt.attributes["dist"].value),
                                       float(cpt.attributes["alt"].value),
                                       cpt.getAttribute("desc"))
        except Exception as err:
            logging.exception(err)
            fid.unlink()
            return False
    for coord in slop.getElementsByTagName('coord'):
        try:
            slopelist.add_coord(slopenum,
                                float(coord.attributes["lat"].value),
                                float(coord.attributes["lon"].value))
        except Exception as err:
            logging.exception(err)
            fid.unlink()
            return False
    fid.unlink()
    return True


def __open_xml(filename, slopelist, slopenum):
    """Open a Cyclomaniac xml file"""
    fid = xml.dom.minidom.parse(filename)
    slopename = fid.getElementsByTagName("name")[0]
    slopename = slopename.childNodes
    slopelist.set_name(slopenum, slopename[0].data)
    samples = fid.getElementsByTagName("samples")[0]
    for sample in samples.getElementsByTagName('sample'):
        for ent in sample.childNodes:
            if ent.localName == "distance":
                dist = ent.childNodes[0].data.replace(',', '.')
            elif ent.localName == "height":
                alt = ent.childNodes[0].data.replace(',', '.')
            elif ent.localName == "label":
                name = ent.childNodes[0].data
        slopelist.add_cp(slopenum, float(dist), float(alt), name)
        name = u""
    fid.unlink()
    return True


def load_gpxfile(xmlfile):
    """Load a gpx file"""
    try:
        if isinstance(xmlfile, str) and xmlfile.strip().startswith('<'):
            fid = xml.dom.minidom.parseString(xmlfile)
        else:
            fid = xml.dom.minidom.parse(xmlfile)
        ptls = fid.getElementsByTagName('trkpt')
        coords = [(el.attributes["lon"].value, el.attributes["lat"].value) for el in ptls]
        cps = distance(coords)
        next_cp = 0
        max_dist = cps[-1][0]
        indices = []
        for i in range(len(cps)):
            dist = cps[i][0]
            if i == 0 or dist > next_cp or i == len(cps)-1:
                indices.append(i)
                next_cp = dist + max_dist/len(cps)
        # Use waypoints to determine the name of a cp.
        waypoints = {}
        for waypoint in fid.getElementsByTagName('wpt'):
            # find the nearest (filtered) point to this waypont
            (wpt_lon, wpt_lat) = point_conversion(waypoint.attributes["lon"].value,
                                                  waypoint.attributes["lat"].value)
            (min_i, min_distance) = (-1, float(3))  # use minimum accetable distance 3 km.
            # print cps
            for i in indices:
                # calculate distance to point i
                (lon, lat) = point_conversion(cps[i][1], cps[i][2])
                dist = haversine(lat, wpt_lat, lon, wpt_lon)
                # print (lon, lat, dist)
                if dist < min_distance:
                    (min_i, min_distance) = (i, dist)
            if min_i >= 0:
                # is found a point near enought to the waypoint: use it.
                waypoints[min_i] = waypoint.getElementsByTagName("name")[0].firstChild.data
        returncp = []
        is_with_alt = True
        alt_count = 0
        for i in indices:
            # get elevation
            (dist, lng, lat) = cps[i]
            sample = ptls[i]
            elelist = sample.getElementsByTagName('ele')
            if len(elelist) > 0:
                ele = elelist[0]
                node = ele.childNodes
                alt = float(node[0].data)
            else:
                is_with_alt = False
                alt = 0
            if (alt == 0):
                alt_count += 1
            if i in waypoints:
                name = waypoints[i]
            else:
                name = u""
            returncp.append((dist, alt, name, lat, lng))
        fid.unlink()
        retfile = SlopeFile(returncp)
        if len(indices) == alt_count:
            is_with_alt = False
        retfile.hasAltitudes = is_with_alt
        return retfile
    except Exception as err:
        logging.exception(err)
        return None


def load_fitlogfile(xmlfile):
    """Load a fitlog file"""
    try:
        if isinstance(xmlfile, str) and xmlfile.strip().startswith('<'):
            fid = xml.dom.minidom.parseString(xmlfile)
        else:
            fid = xml.dom.minidom.parse(xmlfile)
        ptls = fid.getElementsByTagName('pt')
        coords = [(el.attributes["lon"].value, el.attributes["lat"].value) for el in ptls]
        cps = distance(coords)
        next_cp = 0
        max_dist = cps[-1][0]
        indices = []
        for i in range(len(cps)):
            dist = cps[i][0]
            if i == 0 or dist > next_cp or i == len(cps)-1:
                indices.append(i)
                next_cp = dist + max_dist/len(cps)
        returncp = []
        is_with_alt = True
        alt_count = 0
        name = u""
        for i in indices:
            # get elevation
            (dist, lng, lat) = cps[i]
            sample = ptls[i]
            if sample.getAttribute('ele'):
                elev = sample.attributes["ele"].value
                if len(elev) > 0:
                    alt = float(elev)
                else:
                    is_with_alt = False
                    alt = 0
            else:
                is_with_alt = False
                alt = 0
            if (alt == 0):
                alt_count += 1
            returncp.append((dist, alt, name, lat, lng))
        fid.unlink()
        retfile = SlopeFile(returncp)
        if len(indices) == alt_count:
            is_with_alt = False
        retfile.hasAltitudes = is_with_alt
        return retfile
    except Exception as err:
        logging.exception(err)
        return None


def load_kmlfile(xmlfile):
    """Load a kml file"""
    try:
        if isinstance(xmlfile, bytes) and xmlfile.strip().startswith(b'<'):
            fid = xml.dom.minidom.parseString(xmlfile)
        elif isinstance(xmlfile, str) and xmlfile.strip().startswith('<'):
            fid = xml.dom.minidom.parseString(xmlfile)
        else:
            fid = xml.dom.minidom.parse(xmlfile)

        coordlist = []
        altlist = []

        for linstr in fid.getElementsByTagName("LineString"):
            coords = linstr.getElementsByTagName("coordinates")[0]
            node = coords.childNodes[0]

            data = node.data.strip(' \n')
            data = data.replace('\n', ' ')
            coordlist += [el.split(',')[:2] for el in data.split(' ')]

        # if LineString is not present try gx:coord (KML 2.2 ext)
        if len(coordlist) == 0:
            for gxcoord in fid.getElementsByTagName("gx:coord"):
                node = gxcoord.childNodes[0]
                temp = node.data.split(' ')
                coordlist += [temp[:2]]
                altlist += [float(temp[2])]

            if len(coordlist) == 0:
                return SlopeFile([])

        cps = distance(coordlist)

        next_cp = 0
        max_dist = cps[-1][0]
        is_with_alt = len(altlist) >= len(cps)
        returncp = []
        for i in range(len(cps)):
            (dist, lng, lat) = cps[i]
            # create self.num_cps check points
            # print i
            if i == 0 or dist > next_cp or i == len(cps)-1:
                if is_with_alt:
                    # altitude is already in the kml file
                    alt = altlist[i]
                else:
                    # altitude has to be downloaded
                    alt = 0

                returncp.append((dist, alt, "", lat, lng))
                next_cp = dist + max_dist/len(cps)

        fid.unlink()
        retfile = SlopeFile(returncp)
        retfile.hasAltitudes = is_with_alt
        return retfile
    except Exception as err:
        logging.exception(err)
        return None


def load_tcxfile(xmlfile):
    """Load a tcx file"""
    try:
        fid = xml.dom.minidom.parse(xmlfile)
        ptls = fid.getElementsByTagName('Trackpoint')
        coords = []
        is_with_alt = False
        for sample in ptls:
            for ent in sample.childNodes:
                if ent.localName == "Position":
                    # Iterate on child nodes to get latitude and longitude
                    for pos_node in ent.childNodes:
                        if pos_node.localName == "LatitudeDegrees":
                            lat = pos_node.childNodes[0].data
                        elif pos_node.localName == "LongitudeDegrees":
                            lng = pos_node.childNodes[0].data
                elif ent.localName == "AltitudeMeters":
                    alt = float(ent.childNodes[0].data)
                    if (alt != 0):
                        is_with_alt = True
            coords.append((lng, lat, alt))
        dist = 0.0
        (lon_old, lat_old) = point_conversion(coords[0][0], coords[0][1])
        returncp = []
        for i in range(len(coords)):
            (lon_new, lat_new) = point_conversion(coords[i][0], coords[i][1])
            dist += haversine(lat_old, lat_new, lon_old, lon_new)
            lon_old = lon_new
            lat_old = lat_new
            returncp.append((dist, coords[i][2], u"", coords[i][1], coords[i][0]))
        retfile = SlopeFile(returncp)
        logging.info("is with alt " + str(is_with_alt))
        retfile.hasAltitudes = is_with_alt
        return retfile
    except Exception as err:
        logging.exception(err)
        return None

(UNSUPPORTED, CPS, AUTHOR, DATA) = range(4)


def __open_txt(file_, slopelist, slopenum):
    """Open a Cyclomaniac TXT file"""
    if isinstance(file_, str):
        fid = open(file_, encoding='utf-8', mode='r', errors='replace')
    else:
        fid = file_
    # with open(filename, 'rb') as fid:
    mode = CPS
    for line in fid:
        line = line.strip()
        # Find if switching to a different mode.
        if line.startswith('['):
            modename = line.strip('[]').lower()
            # print modename
            if modename == 'author':
                mode = AUTHOR
            elif modename == 'data':
                mode = DATA
            else:
                mode = UNSUPPORTED
            continue

        # mode cps import check point to slope.
        if mode == CPS:
            # format is dist, alt, name, other.
            (dist, alt, name) = line.split(',')[:3]
            (dist, alt) = (float(dist), float(alt))
            name = name.strip('\"')
            slopelist.add_cp(slopenum, dist, alt, name)

        # mode data imports extra information about the slope
        elif mode == DATA:
            try:
                (lhs, rhs) = line.split('=')[:2]
            except:
                continue
            # print lhs +" <- " + rhs
            if lhs == 'IDPercorso':
                slopelist.set_name(slopenum, rhs)
            elif lhs == 'Stato':
                slopelist.set_country(slopenum, rhs)
            elif lhs == 'Url':
                slopelist.set_url(slopenum, rhs)
    fid.close()


def __open_sal(filepathname, slopelist, slopenum):
    """Open a Salitaker SAL file"""
    if isinstance(filepathname, str):
        filename = os.path.split(filepathname)[-1]
        name = filename.split('.')[0]
        slopelist.set_name(slopenum, name)

        fid = open(filepathname, encoding='utf-8', mode='r')
    else:
        fid = filepathname
    # with open(filepathname, 'rb') as fid:
    ncp = fid.readline().strip().split(' ')[4]
    for _ in range(int(ncp)):
        name = fid.readline().strip()
        (dist, alt) = fid.readline().strip().split(' ')
        slopelist.add_cp(slopenum, float(dist), float(alt), name)
    fid.close()


def __open_crp(filepathname, slopelist, slopenum, num_cps=-1):
    """Open  Ciclotour crp
    The parameter num_cps is the number ok cp to import (at most).
    Particular values are:
    0: import every cp
    -1: import the default number of cp"""
    data = b""
    if isinstance(filepathname, str):
        fid = open(filepathname, mode='rb')  # needed byte access not unicode string
    else:
        fid = filepathname
    # with open(filepathname, 'rb') as fid:
    byte = b' '
    while byte:
        byte = fid.read(1)
        data += byte
    fid.close()
    if not data:
        # File not accesible or empty
        return -1

    if data[:2] == b'\x78\xda':
        plaincsv = zlib.decompress(data).decode()
    elif data[:2] == b'HR':
        plaincsv = data.decode()
    else:
        # Invalid file format
        return -1
    # print(plaincsv)
    points = plaincsv.split('***')
    line = points[0].split('\n')   # plaincsv.split('\n')
    # print(line[0])
    next_cp = 0
    max_dist = float(line[-2].split('\t')[2])/100
    lastline = len(line)-3
    if num_cps < 0:
        N = 20  # default value: often looks good.
    elif num_cps == 0:
        N = lastline
    else:
        N = num_cps
    for i in range(lastline):
        fields = line[i+2].split('\t')
        # print fields
        dist = float(fields[2])/100
        # create N check points
        if dist > next_cp or i == lastline or fields[8] != '':
            alt = float(fields[3])
            slopelist.add_cp(slopenum, dist, alt, fields[8])
            next_cp = dist + max_dist/N
    return True

# Function that return kml string from gpx


def gpxtokml(xmlfile):
    """Load a gpx file into kml"""
    if isinstance(xmlfile, str) and xmlfile.strip().startswith('<'):
        fid = xml.dom.minidom.parseString(xmlfile)
    else:
        fid = xml.dom.minidom.parse(xmlfile)
    ptls = fid.getElementsByTagName('trkpt')
    coords = [(el.attributes["lon"].value, el.attributes["lat"].value) for el in ptls]
    kmlstring = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
    kmlstring += "<kml xmlns=\"http://earth.google.com/kml/2.0\" xmlns:atom=\"http://www.w3.org/2005/Atom\">\n"
    kmlstring += "<Document>\n"
    kmlstring += "<LineString><coordinates>"

    for coord in coords:
        (lon, lat) = coord
        kmlstring += lon + "," + lat + " "
    kmlstring += "</coordinates></LineString>\n"
    kmlstring += "</Document>\n</kml>"
    return kmlstring
# class for various formats


class SlopeFile:
    def __init__(self, cps):
        """Load a track file"""
        self.cps = cps
        self.cpToLoad = []
        self.hasAltitudes = False

    def min_dist(self):
        """ Minimum distance of the track """
        return self.cps[0][0]

    def max_dist(self):
        """ Maximum distance of the track """
        return self.cps[-1][0]

    def __len__(self):
        """ Number of points of the track """
        return len(self.cps)

    def newSlope(self, slopelist, slopenum, num, start_dist, end_dist, seltype=0):
        """ Create a new slope in the slopelist"""
        self.setcpToLoad(num, start_dist, end_dist, seltype)
        self.getCoords(slopelist, slopenum, start_dist, end_dist)
        for icp in self.cpToLoad:
            (dist, alt, name, lat, lng) = self.cps[icp]
            normadist = dist - float(start_dist)
            slopelist.add_cp(slopenum, normadist, alt, name)
        return True

    def setcpToLoad(self, num, start_dist, end_dist, seltype):
        """ Set cp index to load"""
        self.cpToLoad = []

        startindex = 0
        endindex = len(self.cps)
        lencps = len(self.cps)

        for i in range(lencps):
            if self.cps[i][0] >= start_dist:
                startindex = i
                break
        for i in range(lencps):
            if self.cps[lencps-i-1][0] <= end_dist:
                endindex = lencps - i
                break

        # print(startindex)
        # print(endindex)

        if seltype == 0:  # using the given number of check points in num
            max_dist = self.max_dist()
            next_cp = 0.0

            if num == -1:
                N = 30
            elif num == 0:
                N = endindex - startindex
            else:
                N = num

            for i in range(startindex, endindex):
                dist = self.cps[i][0]
                # create approximately N check points
                if i == startindex or dist > next_cp or i == endindex-1:
                    self.cpToLoad.append(i)
                    next_cp = dist + max_dist/N
        elif seltype == 1:  # using the passed minimum distance in metres in num
            next_dist = num
            for i in range(startindex, endindex):
                dist = self.cps[i][0]
                if i == startindex or dist > next_dist or i == endindex-1:
                    self.cpToLoad.append(i)
                    next_dist = dist + num
        return True

    def getCoords(self, slopelist, slopenum, start_dist, end_dist):
        """ Get coords from cps to the passed slope"""
        startindex = 0
        endindex = len(self.cps)
        lencps = len(self.cps)

        for i in range(lencps):
            if self.cps[i][0] >= start_dist:
                startindex = i
                break
        for i in range(lencps):
            if self.cps[lencps-i-1][0] <= end_dist:
                endindex = lencps - i
                break

        # print(startindex)
        # print(endindex)
        for i in range(startindex, endindex):
            (dist, alt, name, lat, lng) = self.cps[i]
            slopelist.add_coord(slopenum, float(lat), float(lng))

# vim:sw=4:softtabstop=4:expandtab
